;------------------------------------------------------------------------------
;
; Virtual machine description

; The virtual machine contains registers, a stack, an array of global variables
; and a memory vector:
;
;  registers                       stack               global array
;  ________                                              ________
; |________| 0 (continuation)    |________|             |________| 1
; |________| 1                   |________| 1  top      |________| 2
; |________| 2                   |________| 2           |________| .
; |________| ...                 |________| 3  /|\      |________| .
; |________| nb-regs             |________| ... |       |________| .

(define nb-regs 3)

; Operands of the virtual machine

(define-structure imm val)  ; immediate operand
(define-structure reg num)  ; operand is a register (reg 1 is 1st general reg)
(define-structure stk num)  ; operand is on the stack (topmost is number 1)
(define-structure glob num) ; operand is a global variable (first is number 1)
(define-structure push)     ; to push a value on the the stack
(define-structure pop)      ; to pop a value off the stack
(define-structure indx base offset) ; indexing (offset is in bytes)
(define-structure lbl num)  ; a local label
(define-structure prim name nb-args) ; the entry point of a primitive procedure

(define cont-reg (make-reg 0)) ; the continuation register
(define result-reg (make-reg 1)) ; the result register
(define push-opnd (make-push))
(define pop-opnd (make-pop))

(define-structure slot num) ; operand is a frame slot (gets converted to stk)

;------------------------------------------------------------------------------
;
; Virtual machine implementation (for MC68000)

; Object representation parameters (used for indexing)

(define object-size 4)         ; size of an object
(define proc-offset 2)         ; offset to procedure pointerin a closure object
(define first-closed-offset 6) ; offset to first closed var in a closure object
(define cell-offset 0)         ; offset to value in a cell object

; code output (virtual instr --> M4 macros)

(define (write-instr instr)

  (let ((opcode (first instr)))

    (cond ((eq? opcode 'label) ; ==> LABEL(xxx)
           (write-code "LABEL(" (second instr) ")"))

          ((eq? opcode 'entry-label) ; ==> ENTRY_LABEL(xxx)
           (write-code "ENTRY_LABEL(" (second instr) ")"))

          ((eq? opcode 'if)
           ; ==> IF(src,xxx) / IF_NOT(src,xxx)
           ; ==> IF_pred(src,xxx) / IF_NOT_pred(src,xxx)

           (let ((true (second instr))
                 (cnd (third instr))
                 (src (encode-operand (fourth instr))))
             (write-code tab
                         "IF"
                         (if true '() "_NOT")
                         (cond ((eq? cnd 'true?)      '())
                               ((eq? cnd 'null?)      "_NULL")
                               ((eq? cnd 'fixnum?)    "_FIXNUM")
                               ((eq? cnd 'flonum?)    "_FLONUM")
                               ((eq? cnd 'char?)      "_CHAR")
                               ((eq? cnd 'pair?)      "_PAIR")
                               ((eq? cnd 'hunk?)      "_HUNK")
                               ((eq? cnd 'procedure?) "_PROC")
                               ((eq? cnd 'string?)    "_STRING")
                               ((eq? cnd 'vector?)    "_VECTOR")
                               ((eq? cnd 'symbol?)    "_SYMBOL")
                               (else (error "unknown condition type")))
                         "("
                         src
                         ","
                         (fifth instr)
                         ")")))

          ((or (eq? opcode 'jump) (eq? opcode 'call))
           ; ==> JUMP(src,n) / CALL(src,n)
           ; ==> JUMP(src,n,nb-args) / CALL(src,n,nb-args)
           ; ==> JUMP_CHECK(src,n) / CALL_CHECK(src,n)
           ; ==> JUMP_CHECK(src,n,nb-args) / CALL_CHECK(src,n,nb-args)

           (let ((check? (second instr))
                 (dealloc (third instr))
                 (nb-args (fourth instr))
                 (src (encode-operand (fifth instr))))
             (write-code tab
                         (if (eq? opcode 'jump) "JUMP" "CALL")
                         (if check? "_CHECK(" "(")
                         src
                         ","
                         dealloc
                         (if nb-args (list "," nb-args) '())
                         ")")))

          ((eq? opcode 'enter)
           ; ==> ENTER(nb-args) / ENTER_REST(nb-args)
           ; ==> ENTER_CLO(nb-args) / ENTER_REST_CLO(nb-args)

           (let ((rest? (second instr))
                 (closed? (third instr))
                 (nb-args (fourth instr)))
             (write-code tab
                         "ENTER"
                         (if rest? "_REST" '())
                         (if closed? "_CLO(" "(")
                         nb-args
                         ")")))

          ((eq? opcode 'check-stack) ; ==> CHECK_STACK
           (write-code tab "CHECK_STACK"))

          ((eq? opcode 'make-cell) ; ==> MAKE_CELL(src)
           (let ((src (encode-operand (second instr))))
             (write-code tab "MAKE_CELL(" src ")")))

          ((eq? opcode 'make-closures) ; ==> MAKE_CLOSURES(size1;...;sizen)
           (write-code tab "MAKE_CLOSURES(" (list-of (second instr) ";") ")"))

          ((eq? opcode 'dealloc) ; ==> DEALLOC(n)
           (write-code tab "DEALLOC(" (second instr) ")"))

          ((eq? opcode 'copy) ; ==> COPY(src,dst)

           (let ((src (encode-operand (second instr)))
                 (dst (encode-operand (third instr))))
             (write-code tab "COPY(" src "," dst ")")))

          (else
           (error "unknown instruction type")))))

(define (list-of l sep)
  (let loop ((l1 l) (l2 '()))
    (if (null? l1)
      (reverse l2)
      (loop (after-first l1)
            (cons (if (null? (after-first l1)) '() sep)
                  (cons (first l1)
                        l2))))))

(define (partition l n)
  (if (null? l)
    '()
    (let loop ((l l) (x '()) (y '()) (i (+ n 1)))
      (cond ((null? l)
             (reverse (cons (reverse y) x)))
            ((= i 1)
             (loop (after-first l)
                   (cons (reverse y) x)
                   (list (first l))
                   n))
            (else
             (loop (after-first l)
                   x
                   (cons (first l) y)
                   (- i 1)))))))

; Operand encoding

(define (encode-operand opnd)
  (cond ((imm? opnd)
         (let ((val-repr (object-repr (imm-val opnd))))
           (if val-repr
             (list "imm(" val-repr ")")
             (list "const(" (add-constant (imm-val opnd)) ")"))))
        ((reg? opnd)
         (list "reg(" (reg-num opnd) ")"))
        ((stk? opnd)
         (list "stk(" (stk-num opnd) ")"))
        ((glob? opnd)
         (list "glob(" (glob-num opnd) ")"))
        ((push? opnd)
         "push")
        ((pop? opnd)
         "pop")
        ((indx? opnd)
         (list "indx("
               (encode-operand (indx-base opnd))
               ","
               (indx-offset opnd)
               ")"))
        ((lbl? opnd)
         (list "lbl(" (lbl-num opnd) ")"))
        ((prim? opnd)
         (list "prim(```"
               (prim-name opnd)
               "'''"
               (if (prim-nb-args opnd) (list "," (prim-nb-args opnd)) '())
               ")"))
        (else
         (error "unknown operand type"))))

; Object representation (for non-memory allocated objects)

(define (object-repr obj)
  (cond ((null? obj)    -3)
        ((eq? obj #f)   -1)
        ((eq? obj #t)   -5)
        ((char? obj)    (- (* 2 (char->integer obj)) 131071))
        ((integer? obj) (* obj 8))
        ((real? obj)    (error "real numbers not implemented yet"))
        (else           #f)))

; Machine code output procedures

(define *code-port* '())

(define tab (integer->char 9))

(define (write-code . l)

  (define (write-rest l)
    (cond ((null? l))
          ((pair? l)
           (write-rest (first l))
           (write-rest (after-first l)))
          (else
           (display l *code-port*))))

  (write-rest l)
  (newline *code-port*))

;------------------------------------------------------------------------------
;
; The file compiler

(define (compile-file filename)

  (define (parse expression env)
    (if (define-expr? expression)
      (let ((name (definition-variable expression)))
        (newline) (display "parsing define to ") (write name)
        (list 'define (lookup-var name env)
          (expression->parse-tree (definition-value expression) env)))
      (begin
        (newline) (display "parsing an expression")
        (list 'eval (expression->parse-tree expression env)))))

  (define (read-program port env)
    (let ((expression (read port)))
      (if (eof-object? expression)
        '()
        (let ((x (parse expression env)))
          (cons x (read-program port env))))))

  (define (dump-objects)
    (write-code "|--------------------------------")
    (do ((l (list-tail (queue->list object-queue)
                       (length initial-object-queue))
            (after-first l)))
        ((null? l))
      (write-code)
      (let ((x (first l)))
        (let ((obj (car x)) (num (cdr x)))
          (cond ((pair? obj)
                 (write-code "PAIR_OBJECT(obj" num ","
                             (encode-object (car obj)) ","
                             (encode-object (cdr obj)) ")"))
                ((symbol? obj)
                 (write-code "SYMBOL_OBJECT(obj" num ","
                             (encode-object (symbol->string obj)) ")"))
                ((string? obj)
                 (write-code "STRING_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".byte" tab (list-of l ",")))
                   (partition (map char->integer (string->list obj)) 10))
                 (write-code "END"))
                ((vector? obj)
                 (write-code "VECTOR_OBJECT(obj" num ")")
                 (for-each
                   (lambda (l) (write-code tab ".long" tab (list-of l ",")))
                   (partition (map encode-object (vector->list obj)) 10))
                 (write-code "END"))
                (else
                 (error "unknown object kind")))))))

  (set! *global-env* (make-standard-environment))
  (let ((infile (string-append filename ".scm"))
        (outfile (string-append filename ".s")))
    (let ((input (open-input-file infile)))
      (set! *code-port* (open-output-file outfile))
      (let ((program (read-program input *global-env*)))
        (set! object-counter (make-counter initial-object-counter))
        (set! object-queue (list->queue (append initial-object-queue '())))
        (set! symbol-table initial-symbol-table)
        (newline)
        (compile-procedures program)
        (compile-top-level program)
        (newline)
        (newline) (display "dumping objects")
        (dump-objects)
        (newline))
      (close-input-port input)
      (close-output-port *code-port*))))

(define (encode-object obj)
  (let ((obj-repr (object-repr obj)))
    (if obj-repr
      obj-repr
      (list "obj" (add-object obj)))))

(define *global-env* '())

(define (make-standard-environment)
  (let ((env (make-empty-environment)))
    (set! symbol-table-var (lookup-var 1 env))
    (for-each (lambda (name) (lookup-var name env))
              (map first *primitive-procedures*))
    env))

;------------------------------------------------------------------------------
;
; Runtime context manipulation (i.e. where the variables are, what registers
; are in use, etc.)

; runtime context description: frame-size = number of slots presently allocated
; for the current frame on the stack, slots = list of variables associated with
; each slot (topmost slot first), regs = list of variables contained in each
; register, closed = list of variables which are closed with respect to the
; current procedure.

(define-structure context frame-size slots regs closed)

(define frame-size '())
(define slots      '())
(define regs       '())
(define closed     '())

; special variable used to denote the continuation of a procedure

(define cont-var (make-var 'cont #t set-empty set-empty))
(define cont-var-set (set-singleton cont-var))

; special variable used to denote the pointer to the closed variables

(define closed-var (make-var 'closed #t set-empty set-empty))

; (entry-context proc closed) returns the context in existence upon entry to
; the procedure `proc'.

(define (entry-context proc closed)
  (let ((rest (proc-rest proc))
        (parms (proc-parms proc)))

    (define (locate-parms parms slots)
      (if (or (and (not rest) (<= (length parms) nb-regs))
              (= (length parms) 0))
        (if (null? closed)
          (make-context
            (length slots)
            slots
            (cons cont-var parms)
            closed)
          (make-context
            (+ (length slots) 1)
            (cons closed-var slots)
            (cons cont-var parms)
            closed))
        (locate-parms (cdr parms) (cons (car parms) slots))))

    (locate-parms parms '())))

(define (restore-context context)
  (set! frame-size (context-frame-size context))
  (set! slots      (context-slots context))
  (set! regs       (context-regs context))
  (set! closed     (context-closed context)))

(define (current-context)
  (make-context frame-size slots regs closed))

(define (flush-regs)
  (set! regs '()))

(define (get-var opnd)
  (cond ((reg? opnd)
         (list-ref regs (reg-num opnd)))
        ((slot? opnd)
         (list-ref slots (- frame-size (slot-num opnd))))
        (else
         (error "location must be register or stack slot"))))

(define (put-var opnd new)
  (cond ((reg? opnd)
         (set! regs (replace-nth regs (reg-num opnd) new)))
        ((slot? opnd)
         (set! slots (replace-nth slots (- frame-size (slot-num opnd)) new)))
        (else
         (error "location must be register or stack slot"))))

(define (alloc-slots n)
  (set! frame-size (+ frame-size n))
  (let loop ((i n))
    (if (> i 0)
      (begin
        (set! slots      (cons '() slots))
        (loop (- i 1))))))

(define (dealloc-slots n)
  (set! frame-size (- frame-size n))
  (set! slots      (list-tail slots n)))

(define (push-slot)
  (alloc-slots 1))

(define (pop-slot)
  (dealloc-slots 1))

(define (replace-nth l i v)
  (if (null? l)
    (if (= i 0)
      (list v)
      (cons '() (replace-nth l (- i 1) v)))
    (if (= i 0)
      (cons v (cdr l))
      (cons (car l) (replace-nth (cdr l) (- i 1) v)))))

(define (pos-in-list x l)
  (let ((y (memq x l)))
    (and y (- (length l) (length y)))))

(define (live-vars live)
  (if (not (set-empty? (set-intersection live (list->set closed))))
    (set-adjoin live closed-var)
    live))

(define (dead-slots live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((set-member? (first l) live-v)
             (loop s (after-first l) (- i 1)))
            (else
             (loop (cons i s) (after-first l) (- i 1)))))
    (loop '() slots frame-size)))

(define (live-slots live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((set-member? (first l) live-v)
             (loop (cons i s) (after-first l) (- i 1)))
            (else
             (loop s (after-first l) (- i 1)))))
    (loop '() slots frame-size)))

(define (dead-regs live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((and (set-member? (first l) live-v)
                  (not (memq (first l) slots)))
             (loop s (after-first l) (+ i 1)))
            (else
             (loop (cons i s) (after-first l) (+ i 1)))))
    (loop '() regs 0)))

(define (live-regs live)
  (let ((live-v (live-vars live)))
    (define (loop s l i)
      (cond ((null? l) (list->set (reverse s)))
            ((and (set-member? (first l) live-v)
                  (not (memq (first l) slots)))
             (loop (cons i s) (after-first l) (+ i 1)))
            (else
             (loop s (after-first l) (+ i 1)))))
    (loop '() regs 0)))

(define (lowest-dead-slot live)
  (make-slot (or (lowest (dead-slots live)) (+ frame-size 1))))

(define (highest-live-slot live)
  (make-slot (or (highest (live-slots live)) 0)))

(define (highest set) ; return highest number in the set
  (if (set-empty? set) #f (apply max (set->list set))))

(define (lowest set) ; return lowest number in the set
  (if (set-empty? set) #f (apply min (set->list set))))

(define (above set n) ; return numbers above n in the set
  (set-keep (lambda (x) (> x n)) set))
  
(define (below set n) ; return numbers below n in the set
  (set-keep (lambda (x) (< x n)) set))
  
(define (var->opnd* var)
  (let ((n (pos-in-list var *global-env*)))
    (if n
      (make-glob n)
      (let ((n (pos-in-list var regs)))
        (if n
          (make-reg n)
          (let ((n (pos-in-list var slots)))
            (if n
              (make-slot (- frame-size n))
              (let ((n (pos-in-list var closed)))
                (if n
                  (make-indx (var->opnd* closed-var)
                             (+ first-closed-offset (* object-size n)))
                  (let ((x (assq var constant-vars)))
                    (if x
                      (cdr x)
                      (error "variable is not accessible"))))))))))))

(define (var->opnd var)
  (if (and (not (global? var)) (mutable? var))
    (make-indx (var->opnd* var) cell-offset)
    (var->opnd* var)))

;------------------------------------------------------------------------------

(define-structure counter value)

(define instr-queue '())
(define proc-queue '())
(define const-queue '())
(define label-counter '())
(define object-counter '())
(define object-queue '())
(define constant-vars '())
(define known-procs '())
(define symbol-table '())
(define symbol-table-var '())

;------------------------------------------------------------------------------

(define (compile-top-level program)

  (define (compile-list l)
    (if (not (null? l))
      (begin
        (cond ((eq? (caar l) 'define)
               (let* ((var (cadar l))
                      (name (var-name var))
                      (value (caddar l)))
                 (newline) (display "compiling define to ") (write name)
                 (let ((opnd (if (proc? value)
                               (make-prim name #f)
                               (gen-node value cont-var-set 'need))))
                   (emit-copy opnd (var->opnd var) cont-var-set))))
              ((eq? (caar l) 'eval)
               (let ((value (cadar l)))
                 (newline) (display "compiling an expression")
                 (gen-node value cont-var-set
                   (if (null? (after-first l)) 'tail 'side)))))
        (compile-list (cdr l)))))

  (newline)

  (set! instr-queue (queue-empty))
  (set! proc-queue (queue-empty))
  (set! const-queue (queue-empty))
  (set! label-counter (make-counter 0))
  (set! constant-vars '())
  (set! known-procs '())

  (restore-context (make-context 0 '() (list cont-var) '()))

  (compile-list program)

  (do () ((queue-empty? proc-queue))
    (apply gen-proc (queue-get! proc-queue)))

  (write-code "|--------------------------------")
  (write-code)
  (write-code "DEFINE_PRIMITIVE(`program-top-level')")

  (do ((l (cons (list 'copy (make-imm symbol-table)
                            (var->opnd symbol-table-var))
                (queue->list instr-queue))
          (after-first l)))
      ((null? l))
    (write-instr (first l)))

  (do ((l (queue->list const-queue) (after-first l)))
      ((null? l))
    (write-code "CONST(obj" (add-object (first l)) ")"))

  (write-code "END")
  (write-code))

;------------------------------------------------------------------------------

(define (compile-procedures l)
  (if (not (null? l))
    (begin
      (if (eq? (caar l) 'define)
        (let* ((var (cadar l))
               (name (var-name var))
               (value (caddar l)))
          (if (proc? value)
            (begin
              (newline) (display "compiling procedure ") (write name)
              (compile-procedure name value)))))
    (compile-procedures (cdr l)))))

;------------------------------------------------------------------------------

(define (compile-procedure name proc)

  (set! instr-queue (queue-empty))
  (set! proc-queue (queue-empty))
  (set! const-queue (queue-empty))
  (set! label-counter (make-counter 0))
  (set! constant-vars '())
  (set! known-procs '())

  (gen-body proc '())
  (do () ((queue-empty? proc-queue))
    (apply gen-proc (queue-get! proc-queue)))

  (write-code (if compile-primitives "DEFINE_PRIMITIVE" "DEFINE_PROCEDURE")
    "(`" name "'," (length (proc-parms proc))
    (if (proc-rest proc) ",REST" '())
    ")")

  (do ((l (queue->list instr-queue) (after-first l)))
      ((null? l))
    (write-instr (first l)))

  (do ((l (queue->list const-queue) (after-first l)))
      ((null? l))
    (write-code "CONST(obj" (add-object (first l)) ")"))

  (write-code "END")
  (write-code))

(define (new-label)
  (counter-value-set! label-counter (+ (counter-value label-counter) 1))
  (counter-value label-counter))

(define (add-constant val)
  (let ((n (pos-in-list val (queue->list const-queue))))
    (if n
      (+ n 1)
      (begin      
        (queue-put! const-queue val)
        (queue-length const-queue)))))

(define (add-object obj)
  (let ((x (assq obj (queue->list object-queue))))
    (if x
      (cdr x)
      (begin
        (if (symbol? obj)
          (set! symbol-table (cons obj symbol-table)))
        (counter-value-set! object-counter
          (+ (counter-value object-counter) 1))
        (queue-put! object-queue
          (cons obj (counter-value object-counter)))
        (counter-value object-counter)))))

(define (add-constant-var var opnd)
  (set! constant-vars (cons (cons var opnd) constant-vars)))

(define (constant-var? var)
  (assq var constant-vars))

(define (not-constant-var? var)
  (not (constant-var? var)))

(define (add-known-proc label proc)
  (set! known-procs (cons (cons label proc) known-procs)))

(define assume-standard-prims #t)

;------------------------------------------------------------------------------

; generate code for a procedure

(define (gen-proc proc label closed)
  (emit-entry-label label)
  (emit-enter (proc-rest proc) (not (null? closed)) (length (proc-parms proc)))
  (emit-label (+ label 1))
  (emit-check-stack)
  (gen-body proc closed))

(define (schedule-gen-proc proc closed-list)
  (let* ((label1 (new-label))  ; arg check entry point
         (label2 (new-label))) ; no arg check entry point
    (queue-put! proc-queue (list proc label1 closed-list))
    (make-lbl label1)))
 
(define (prepare-closure proc-var proc-label closed-list live)
  (emit-copy proc-label
             (make-indx (var->opnd proc-var) proc-offset)
             (set-union (list->set closed-list) (set-adjoin live proc-var)))
  (let loop ((l closed-list) (i 0))
    (if (not (null? l))
      (begin
        (emit-copy (var->opnd* (first l))
                   (make-indx (var->opnd proc-var)
                              (+ first-closed-offset (* object-size i)))
                   (set-union (list->set (after-first l))
                              (set-adjoin live proc-var)))
        (loop (after-first l) (+ i 1))))))

;------------------------------------------------------------------------------

; generate code for the body of a procedure

(define (gen-body proc closed)
  (let ((live (set-adjoin (free-variables (proc-body proc)) cont-var)))
    (restore-context (entry-context proc closed))
    (for-each
      (lambda (parm)
        (if (mutable? parm) (emit-make-cell (var->opnd* parm) live)))
      (proc-parms proc))
    (gen-node (proc-body proc) cont-var-set 'tail)))

;------------------------------------------------------------------------------

; generate code for an expression

(define (gen-node node live why)

  (cond ((root? node)
         (error "Root nodes should not appear inside parse trees"))

        ((uns? node)
         (gen-return result-reg why))

        ((cst? node)
         (gen-return
           (let ((x (primitive-procedure->name (cst-val node))))
             (if x (make-prim x #f) (make-imm (cst-val node))))
           why))

        ((ref? node)
         (gen-return
           (cond ((eq? why 'side)
                  result-reg)
                 ((and assume-standard-prims
                       (global? (ref-var node))
                       (primitive-procedure (var-name (ref-var node))))
                  (make-prim (var-name (ref-var node)) #f))
                 (else
                  (var->opnd (ref-var node))))
           why))

        ((set? node)
         (let* ((src (gen-node (set-val node)
                               (set-adjoin live (set-var node))
                               'keep))
                (dst (var->opnd (set-var node))))
           (emit-copy src dst live))
         (gen-return result-reg why))

        ((cond? node)
         (gen-cond (cond-pre node) (cond-con node) (cond-alt node) live why))

        ((proc? node)
         (let* ((closed (not-constant-closed-vars node))
                (closed-list (set->list closed))
                (proc-label (schedule-gen-proc node closed-list)))
           (let ((opnd
                  (if (null? closed-list)
                    (begin
                      (add-known-proc (lbl-num proc-label) node)
                      proc-label)
                    (begin
                      (emit-dealloc (- frame-size
                                       (slot-num (highest-live-slot
                                                   (set-union closed live)))))
                      (emit-make-closures (list (length closed-list)))
                      (let ((slot (make-slot frame-size))
                            (var (make-arg-var 'closure)))
                        (put-var slot var)
                        (prepare-closure var proc-label closed-list live)
                        slot)))))
             (gen-return opnd why))))

        ((call? node)
         (gen-call (call-oper node) (call-args node) live why))

        (else
         (error "Unknown parse tree node type"))))

(define (gen-return opnd why)
  (cond ((eq? why 'tail)
         (emit-copy opnd result-reg cont-var-set)
         (emit-jump #f 0 #f (var->opnd cont-var) set-empty))
        (else
         opnd)))

(define (not-constant-closed-vars proc)
  (set->list (set-keep not-constant-var? (closed-variables proc))))

;------------------------------------------------------------------------------

(define (gen-cond pre con alt live why)

  (define (gen-test true test pre)

    (if (not (eq? why 'tail)) ; could be removed if merge-contexts was smarter.
      ; save regs on stack if they contain values needed after the conditional
      (save-regs (live-regs live)
                 (set-union live
                            (free-variables pre)
                            (free-variables con)
                            (free-variables alt))))

    (let* ((liv (set-union live (free-variables con) (free-variables alt)))
           (opnd (gen-node pre liv 'need)))
      (let ((label1 (new-label)))
        (emit-if true test opnd label1 liv)
        (let ((pre-context (current-context)))
          (if (eq? why 'tail)

            (begin
              (gen-node con live why)
              (restore-context pre-context)
              (emit-label label1)
              (gen-node alt live why))

            (begin

              (let ((con-opnd (gen-node con live why)))
                (if (and (set-member? 1 (live-regs live))
                         (not (and (reg? con-opnd)
                                   (= (reg-num con-opnd) 1))))
                  (save-opnd result-reg live))
                (emit-copy con-opnd result-reg live))

              (let ((con-instr-queue instr-queue)
                    (con-context (current-context)))
                (set! instr-queue (queue-empty))
                (restore-context pre-context)

                (let ((alt-opnd (gen-node alt live why)))
                  (if (and (set-member? 1 (live-regs live))
                           (not (and (reg? alt-opnd)
                                     (= (reg-num alt-opnd) 1))))
                    (save-opnd result-reg live))
                  (emit-copy alt-opnd result-reg live))

                (let ((alt-instr-queue instr-queue)
                      (alt-context (current-context)))
                  (set! instr-queue con-instr-queue)
                  (restore-context con-context)

                  (merge-contexts alt-context alt-instr-queue live)

                  (if (not (queue-empty? alt-instr-queue))
                    (let ((label2 (new-label)))
                      (emit-jump #f frame-size #f (make-lbl label2) live)
                      (emit-label label1)
                      (for-each put-instr (queue->list alt-instr-queue))
                      (emit-label label2))
                    (emit-label label1))

                  result-reg))))))))

  (define (gen-cond* true pre)
    (if (and assume-standard-prims
             (call? pre)
             (ref? (call-oper pre))
             (global? (ref-var (call-oper pre)))
             (= (length (call-args pre)) 1)
             (memq (var-name (ref-var (call-oper pre)))
               '(not null? pair? char? procedure? string? vector? symbol?)))
      (if (eq? (var-name (ref-var (call-oper pre))) 'not)
        (gen-cond* (not true) (first (call-args pre)))
        (gen-test true
                  (var-name (ref-var (call-oper pre)))
                  (first (call-args pre))))
      (gen-test true 'true? pre)))

  (gen-cond* #f pre))

(define (merge-contexts alt-context alt-instr-queue live)
;(write (set-map var-name slots))
;(write (set-map var-name regs))
;(write (set-map var-name (context-slots alt-context)))
;(write (set-map var-name (context-regs alt-context)))
;(newline)
  (let ((live-v (live-vars live)))
    (if (or (not (same-var-list?
                   (reverse (context-slots alt-context))
                   (reverse slots)
                   live-v))
            (not (same-var-list?
                   (context-regs alt-context)
                   regs
                   (set-difference live-v (list->set slots)))))
      (error "context merging for conditionals not yet implemented")
      (let ((dealloc-to (slot-num (highest-live-slot live))))
        (emit-dealloc (- frame-size dealloc-to))
        (let ((con-instr-queue instr-queue)
              (con-context (current-context)))
          (set! instr-queue alt-instr-queue)
          (restore-context alt-context)
          (emit-dealloc (- frame-size dealloc-to))
          (set! instr-queue con-instr-queue)
          (restore-context con-context)
          (set! regs (cancel-diff-vars regs (context-regs alt-context))))))))

(define (same-var-list? l1 l2 live-v)
  (cond ((null? l1)
         (or (null? l2) (same-var-list? l2 l1 live-v)))
        ((null? l2)
         (and (not (set-member? (first l1) live-v))
              (same-var-list? (after-first l1) '() live-v)))
        ((eq? (first l1) (first l2))
         (same-var-list? (after-first l1) (after-first l2) live-v))
        (else
         (and (not (set-member? (first l1) live-v))
              (not (set-member? (first l2) live-v))))))

(define (cancel-diff-vars l1 l2)
  (cond ((null? l1)
         '())
        ((null? l2)
         '())
        ((eq? (first l1) (first l2))
         (cons (first l1)
               (cancel-diff-vars (after-first l1) (after-first l2))))
        (else
         (cons '()
               (cancel-diff-vars (after-first l1) (after-first l2))))))

;------------------------------------------------------------------------------

; generate code for a procedure call

(define (gen-call oper args live why)
  (let ((nb-args (length args)))

    (if (and (proc? oper) ; applying a lambda-expr is like a 'let' or 'letrec'
             (not (proc-rest oper))
             (= (length (proc-parms oper)) nb-args))

      (gen-let (proc-parms oper) args (proc-body oper) live why)

      (let ((eval-order (arg-eval-order oper args))
            (live-after (if (eq? why 'tail) (set-remove live cont-var) live)))

        ; save regs on stack if they contain values needed after the call
        (save-regs (live-regs live-after) (args-live-vars live eval-order))

        (let ((frame-start (slot-num (highest-live-slot live-after))))

          (let loop ((l eval-order) (x live) (stacked '()) (oper-var '()))
            (if (null? l)

              (let* ((oper-opnd (var->opnd oper-var))
                     (exist-entry-point
                       (or (and (prim? oper-opnd)
                                (primitive-entry-point (prim-name oper-opnd)
                                                       nb-args))
                           (and (lbl? oper-opnd)
                                (let ((x (assq (lbl-num oper-opnd) known-procs)))
                                  (and x
                                       (let ((proc (cdr x)))
                                         (and (not (proc-rest proc))
                                              (= (length (proc-parms proc))
                                                 nb-args))))))))
                     (nb-args-check
                       (if exist-entry-point #f nb-args))
                     (opnd
                       (cond ((and exist-entry-point (prim? oper-opnd))
                              (make-prim (prim-name oper-opnd) nb-args))
                             ((and exist-entry-point (lbl? oper-opnd))
                              (make-lbl (+ (lbl-num oper-opnd) 1)))
                             (else
                              oper-opnd))))

                (do ((p stacked (after-first p)))
                    ((null? p))
                  (let ((slot (first (first p)))
                        (reg (second (first p)))
                        (liv (third (first p)))
                        (var (fourth (first p))))
                    (emit-copy slot reg liv)
                    (put-var reg var)))

                (if (eq? why 'tail)
                  (let ((liv (set-remove x cont-var)))
                    (emit-copy (var->opnd cont-var) cont-reg liv)
                    (emit-jump (not (or (prim? opnd) (lbl? opnd)))
                      (+ frame-start (max 0 (- nb-args nb-regs)))
                      nb-args-check opnd (set-remove liv oper-var)))
                  (emit-call (not (or (prim? opnd) (lbl? opnd)))
                    (+ frame-start (max 0 (- nb-args nb-regs)))
                    nb-args-check opnd (set-remove x oper-var)))

                (dealloc-slots (- frame-size frame-start))
                (flush-regs)
                result-reg)

              (let* ((arg (car (first l)))
                     (pos (cdr (first l)))
                     (needed (args-live-vars x (after-first l)))
                     (opnd (gen-node arg needed 'need))
                     (var (make-arg-var pos)))
                (if (= pos 0) ; operator
                  (if (and (ref? arg)
                           (not (or (imm? opnd) (prim? opnd) (lbl? opnd))))
                    (loop (after-first l)
                          (set-adjoin x (ref-var arg))
                          stacked
                          (ref-var arg))
                    (begin
                      (save-arg opnd var needed)
                      (loop (after-first l)
                            (set-adjoin x var)
                            stacked
                            var)))
                    (if (<= pos (- nb-args nb-regs)) ; arg passed on stack
                       (let ((slot (make-slot (+ frame-start pos))))
                         (if (set-member? (slot-num slot) (live-slots needed))
                           (save-opnd slot
                             (if (slot? opnd)
                               (set-adjoin needed (get-var opnd))
                               needed)))
                         (emit-copy opnd slot needed)
                         (put-var slot var)
                         (loop (after-first l)
                               (set-adjoin x var)
                               stacked
                               oper-var))
                       (let ((reg (make-reg (- pos (max 0 (- nb-args nb-regs))))))
                         (if (every-trivial? (after-first l))
                           (begin
                             (if (set-member? (reg-num reg) (live-regs needed))
                               (save-opnd reg
                                 (if (slot? opnd)
                                   (set-adjoin needed (get-var opnd))
                                   needed)))
                             (emit-copy opnd reg needed)
                             (put-var reg var)
                             (loop (after-first l)
                                   (set-adjoin x var)
                                   stacked
                                   oper-var))
                           (let ((slot (make-slot (+ frame-size 1))))
                             (emit-copy opnd slot needed)
                             (put-var slot var)
                             (loop (after-first l)
                                   (set-adjoin x var)
                                   (cons (list slot reg x var) stacked)
                                   oper-var))))))))))))))

(define (save-opnd opnd live)
  (let ((slot (lowest-dead-slot live)))
    (emit-copy opnd slot live)
    (put-var slot (get-var opnd))))

(define (save-regs regs live)
  (for-each (lambda (i) (save-opnd (make-reg i) live)) (set->list regs)))

(define (every-trivial? l)
  (or (null? l)
      (and (trivial? (car (first l)))
           (every-trivial? (after-first l)))))

(define (trivial? node)
  (or (uns? node)
      (cst? node)
      (ref? node)
      (and (set? node) (trivial? (set-val node)))))

(define (make-arg-var pos)
  (make-var pos #t set-empty set-empty))

(define (arg-eval-order oper args)
  (let ((nb-args (length args)))

    (define (loop1 args pos first-args)
      (if (and (not (null? args)) (<= pos (- nb-args nb-regs)))
        (loop1 (after-first args)
               (+ pos 1)
               (cons (cons (first args) pos) first-args))
        (append (reverse first-args) (loop2 args pos '() '()))))

    (define (loop2 args pos part1 part2)
      (cond ((null? args)
             (if (trivial? oper)
                (append part1 part2 (list (cons oper 0)))
                (append (cons (cons oper 0) part1) part2)))
            ((trivial? (first args))
             (loop2 (after-first args)
                    (+ pos 1)
                    part1
                    (cons (cons (first args) pos) part2)))
            (else
             (loop2 (after-first args)
                    (+ pos 1)
                    (cons (cons (first args) pos) part1)
                    part2))))

    (loop1 args 1 '())))

(define (args-live-vars live order)
  (if (null? order)
    live
    (args-live-vars (set-union live (free-variables (car (first order))))
                    (after-first order))))

;------------------------------------------------------------------------------

; generate code for a 'let' or 'letrec'

(define (gen-let vars vals node live why)
  (let ((var-val-map (map cons vars vals))
        (var-set (list->set vars))
        (all-live (apply set-union live
                    (free-variables node)
                    (map free-variables vals))))

    (define (var->val var) (cdr (assq var var-val-map)))

    (define (proc-var? var) (proc? (var->val var)))

    (define (non-mutable? var) (not (mutable? var)))

    (let ((proc-vars (set-keep proc-var? var-set)))

      (let loop1 ((const-proc-vars (set-keep non-mutable? proc-vars)))

        (define (closed-vars var)
          (set-difference (not-constant-closed-vars (var->val var))
                          const-proc-vars))

        (define (no-closed-vars? var)
          (set-empty? (closed-vars var)))

        (define (closed-vars? var)
          (not (no-closed-vars? var)))

        (let ((new-const-proc-vars (set-keep no-closed-vars? const-proc-vars)))
          (if (not (set-equal? new-const-proc-vars const-proc-vars))
            (loop1 new-const-proc-vars)

            (let ((clo-vars (set-keep closed-vars? proc-vars)))

              (for-each
                (lambda (proc-var)
                  (let ((label (schedule-gen-proc (var->val proc-var) '())))
                    (add-known-proc (lbl-num label) (var->val proc-var))
                    (add-constant-var proc-var label)))
                (set->list const-proc-vars))

              (if (not (set-empty? clo-vars))
                (let ((clo-vars-list (set->list clo-vars)))
                  (emit-dealloc
                    (- frame-size (slot-num (highest-live-slot all-live))))
                  (emit-make-closures (map length (map closed-vars clo-vars-list)))
                  (let loop2 ((l clo-vars-list)
                              (i (+ (- frame-size (length clo-vars-list)) 1)))
                    (if (not (null? l))
                      (let ((var (first l))
                            (slot (make-slot i)))
                        (put-var slot var)
                        (if (mutable? var) (emit-make-cell slot all-live))
                        (loop2 (after-first l) (+ i 1)))))
                  (for-each
                    (lambda (proc-var)
                      (let ((closed-list (set->list (closed-vars proc-var))))
                        (prepare-closure proc-var
                                         (schedule-gen-proc (var->val proc-var)
                                                            closed-list)
                                         closed-list
                                         all-live)))
                    clo-vars-list)))

              (let ((other-vars-list
                      (set->list
                        (set-keep (lambda (var)
                                    (and (not (set-member? var const-proc-vars))
                                         (not (set-member? var clo-vars))))
                                  vars)))
                    (liv (set-union live (free-variables node))))
                (let loop3 ((vars* other-vars-list))
                  (if (not (null? vars*))
                    (let* ((var (first vars*))
                           (val (var->val var))
                           (needed (vals-live-vars liv
                                     (map var->val (after-first vars*)))))
                      (if (var-useless? var)
                        (gen-node val needed 'side)
                        (begin
                          (save-val (gen-node val needed 'need) var needed)
                          (if (mutable? var)
                            (emit-make-cell (var->opnd* var) needed))))
                      (loop3 (after-first vars*)))))
                (gen-node node live why)))))))))

(define (save-arg opnd var live)
  (if (glob? opnd)
    (add-constant-var var opnd)
    (save-val opnd var live)))

(define (save-val opnd var live)
  (cond ((or (imm? opnd) (prim? opnd) (lbl? opnd))
         (add-constant-var var opnd))
        ((and (reg? opnd)
              (not (set-member? (reg-num opnd) (live-regs live))))
         (put-var opnd var))
        ((and (slot? opnd)
              (not (set-member? (slot-num opnd) (live-slots live))))
         (put-var opnd var))
        (else
         (let ((slot (lowest-dead-slot live)))
           (emit-copy opnd slot live)
           (put-var slot var)))))

(define (var-useless? var)
  (and (set-empty? (var-refs var))
       (set-empty? (var-sets var))))

(define (vals-live-vars live vals)
  (if (null? vals)
    live
    (vals-live-vars (set-union live (free-variables (first vals)))
                    (after-first vals))))

;------------------------------------------------------------------------------

(define (emit-label num)
  (put-instr (list 'label num)))

(define (emit-entry-label num)
  (put-instr (list 'entry-label num)))

(define (emit-check-stack)
  (put-instr '(check-stack)))

(define (emit-enter rest? closed? nb-args)
  (put-instr (list 'enter rest? closed? nb-args)))

(define (emit-if true cnd src label live)
  (let ((src-opnd (emit-source src live)))
    (put-instr (list 'if true cnd src-opnd label))))

(define (emit-jump check? dealloc-to nb-args src live)
  (let* ((src-opnd (emit-source src live))
         (dealloc (- frame-size dealloc-to)))
    (dealloc-slots dealloc)
    (put-instr (list 'jump check? dealloc nb-args src-opnd))))

(define (emit-call check? dealloc-to nb-args src live)
  (let* ((src-opnd (emit-source src live))
         (dealloc (- frame-size dealloc-to)))
    (dealloc-slots dealloc)
    (put-instr (list 'call check? dealloc nb-args src-opnd))))

(define (emit-make-cell src live)
  (let ((src-opnd (emit-source src live)))
    (put-instr (list 'make-cell src-opnd))))

(define (emit-make-closures sizes)
  (alloc-slots (length sizes))
  (put-instr (list 'make-closures sizes)))

(define (emit-dealloc n)
  (if (> n 0)
    (begin
      (dealloc-slots n)
      (put-instr (list 'dealloc n)))))

(define (emit-copy src dst live)
  (if (not (equal? src dst))
    (let* ((dst-slot (slot-num-in dst))
           (src-opnd (emit-operand src live dst-slot))
           (dst-opnd (emit-operand dst live dst-slot)))
      (put-instr (list 'copy src-opnd dst-opnd)))))

(define (slot-num-in opnd)
  (cond ((slot? opnd) (slot-num opnd))
        ((indx? opnd) (slot-num-in (indx-base opnd)))
        (else         -1)))

(define (replace-slot-in opnd new)
  (cond ((slot? opnd) new)
        ((indx? opnd) (make-indx (replace-slot-in (indx-base opnd) new)
                                 (indx-offset opnd)))
        (else         (error "operand does not contain a slot reference"))))

(define (emit-source opnd live)
  (emit-operand opnd live 0))

(define (emit-operand opnd live n)
  (if (and (slot? opnd) (> (slot-num opnd) frame-size))
    (begin (push-slot) push-opnd)
    (let ((opnd-slot (slot-num-in opnd)))
      (if (< opnd-slot 0)
        opnd
        (let ((highest-live-slot (max (or (highest (live-slots live)) 0) n)))
          (if (>= opnd-slot highest-live-slot)
            (emit-dealloc (- frame-size opnd-slot)))
          (replace-slot-in opnd
            (if (> opnd-slot highest-live-slot)
              (begin (pop-slot) pop-opnd)
              (make-stk (+ (- frame-size opnd-slot) 1)))))))))

(define (put-instr instr)
;(newline) (display "--- ") (write instr)
  (queue-put! instr-queue instr))

;------------------------------------------------------------------------------
