;------------------------------------------------------------------------------
;
; PARSE TREE manipulation package
;
; This package contains procedures to construct the parse tree of a Scheme
; expression and manipulate the parse tree.
;
;------------------------------------------------------------------------------

; Definition of the structures found in the parse tree.

(define-structure var ; structure that represents variables
  name                ; symbol that denotes the variable
  bound               ; procedure node that binds the variable (#f if global)
  refs                ; set of nodes that reference this variable
  sets                ; set of nodes that assign a value to this variable
)

; These structures define the nodes associated to expressions.

(define-structure node ; information common to all nodes
  parent               ; the node of which this node is a child
  children             ; list of parse-trees of the sub-expressions
  mark                 ; to keep track of the nodes that have been traversed
  fv                   ; set of free variables contained in this expression
)

(define-node root ; node that represents the root of the parse tree
)                 ; (this kind of node is only used when all nodes must have
                  ; a parent)

(define-node uns ; node that represents an unspecified value
)

(define-node cst ; node that represents constants
  val            ; value of the constant
)

(define-node ref ; node that represents variable references
  var            ; the variable which is referenced
)

(define-node set ; node that represents assignments (i.e. `set!' special forms)
  var            ; the variable which is assigned a value
)

(define-node cond ; node that represents conditionals (i.e. `if' special forms)
)

(define-node proc ; node that represents procedures (i.e. lambda-expressions)
  rest            ; #t if the last parameter is a rest parameter
  parms           ; the list of parameter variables in order
)

(define-node call ; node that represents procedure calls
)

;------------------------------------------------------------------------------

; Procedures to create parse tree nodes and extract sub-nodes.

(define (new-root body)
  (let ((node (make-root #f (list body) #f #t)))
    (node-parent-set! body node)
    node))

(define (root-body x)
  (if (root? x) (first (node-children x)) (error "'root' node expected")))

(define (new-uns)
  (make-uns #f '() #f #t))

(define (new-cst val)
  (make-cst #f '() #f #t val))

(define (new-ref var)
  (let ((node (make-ref #f '() #f #t var)))
    (var-refs-set! var (set-adjoin (var-refs var) node))
    node))

(define (new-set var val)
  (let ((node (make-set #f (list val) #f #t var)))
    (var-sets-set! var (set-adjoin (var-sets var) node))
    (node-parent-set! val node)
    node))

(define (set-val x)
  (if (set? x) (first (node-children x)) (error "'set' node expected")))

(define (new-cond pre con alt)
  (let ((node (make-cond #f (list pre con alt) #f #t)))
    (node-parent-set! pre node)
    (node-parent-set! con node)
    (node-parent-set! alt node)
    node))

(define (cond-pre x)
  (if (cond? x) (first (node-children x)) (error "'cond' node expected")))

(define (cond-con x)
  (if (cond? x) (second (node-children x)) (error "'cond' node expected")))

(define (cond-alt x)
  (if (cond? x) (third (node-children x)) (error "'cond' node expected")))

(define (new-proc rest parms body)
  (let ((node (make-proc #f (list body) #f #t rest parms)))
    (for-each (lambda (x) (var-bound-set! x node)) parms)
    (node-parent-set! body node)
    node))

(define (proc-body x)
  (if (proc? x) (first (node-children x)) (error "'proc' node expected")))

(define (new-parameters parms)
  (cond ((null? parms)
         '())
        ((pair? parms)
         (cons (make-var (first parms) #f set-empty set-empty)
               (new-parameters (after-first parms))))
        (else
         (list (make-var parms #f set-empty set-empty)))))

(define (rest-param? parms)
  (if (pair? parms)
    (rest-param? (after-first parms))
    (not (null? parms))))

(define (new-call oper args)
  (let ((node (make-call #f (cons oper args) #f #t)))
    (node-parent-set! oper node)
    (for-each (lambda (x) (node-parent-set! x node)) args)
    node))

(define (call-oper x)
  (if (call? x) (first (node-children x)) (error "'call' node expected")))

(define (call-args x)
  (if (call? x) (after-first (node-children x)) (error "'call' node expected")))

(define (new-variables vars)
  (if (null? vars)
    '()
    (cons (make-var (first vars) #f set-empty set-empty)
          (new-variables (after-first vars)))))

(define (new-conj pre alt)
  (new-call
    (let* ((parms (new-parameters '(temp)))
           (temp (first parms)))
      (new-proc #f parms (new-cond (new-ref temp) alt (new-ref temp))))
    (list pre)))

(define (new-disj pre alt)
  (new-call
    (let* ((parms (new-parameters '(temp)))
           (temp (first parms)))
      (new-proc #f parms (new-cond (new-ref temp) (new-ref temp) alt)))
    (list pre)))

(define (new-disj-call pre oper alt)
  (new-call
    (let* ((parms (new-parameters '(temp)))
           (temp (first parms)))
      (new-proc #f parms (new-cond (new-ref temp)
                                   (new-call oper (list (new-ref temp)))
                                   alt)))
    (list pre)))

(define (new-seq before after)
  (new-call (new-proc #f (new-parameters '(temp)) after) (list before)))

(define (make-empty-environment)
  (list (make-var))) ; dummy entry in the environment (for global var.)

(define (lookup-var name env)
  (cond ((eq? (var-name (first env)) name)
         (first env))
        ((null? (after-first env)) ; it's a global variable
         (set-cdr! env (list (make-var name #f set-empty set-empty)))
         (second env))
        (else
         (lookup-var name (after-first env)))))

(define (new-env vars env)
  (append vars env))

(define (replace-node! node new-node)
  (define (invalidate-fv node)
    (if node
      (begin
        (node-fv-set! node #t)
        (invalidate-fv (node-parent node)))))
  (invalidate-fv (node-parent node))
  (let ((x (memq node (node-children (node-parent node)))))
    (if x
      (begin
        (set-car! x new-node)
        (node-parent-set! new-node (node-parent node))
        (node-parent-set! node #f))
      (error "Incoherent parse-tree, parent does not contain the node"))))

(define (delete-node! node)
  (define (delete-node node)
    (cond ((root? node)
           (error "Can not delete a root node"))
          ((uns? node))
          ((cst? node))
          ((ref? node)
           (let ((var (ref-var node)))
             (var-refs-set! var (set-remove (var-refs var) node))))
          ((set? node)
           (delete-node (set-val node))
           (let ((var (set-var node)))
             (var-sets-set! var (set-remove (var-sets var) node))))
          ((cond? node)
           (delete-node (cond-pre node))
           (delete-node (cond-con node))
           (delete-node (cond-alt node)))
          ((proc? node)
           (delete-node (proc-body node)))
          ((call? node)
           (delete-node (call-oper node))
           (for-each delete-node (call-args node)))
          (else
           (error "Unknown parse tree node type"))))
  (if (node-parent node)
    (error "Node is still attached to a parse tree")
    (delete-node node)))

(define (extract-node! node)
  (replace-node! node (new-uns))
  node)

;------------------------------------------------------------------------------

; Procedures to get variable classes from nodes.

(define (free-variables node) ; set of free variables used in the expression
  (if (eq? (node-fv node) #t)
    (let ((x (apply set-union (map free-variables (node-children node)))))
      (node-fv-set! node
        (cond ((ref? node)
               (set-adjoin x (ref-var node)))
              ((set? node)
               (set-adjoin x (set-var node)))
              ((proc? node)
               (set-difference x (list->set (proc-parms node))))
              ((and (call? node) (proc? (call-oper node)))
               (set-difference x (list->set (proc-parms (call-oper node)))))
              (else
               x)))))
  (node-fv node))

(define (closed-variables proc) ; set of closed variables of a procedure
  (set-keep bound? (free-variables proc)))

(define (bound-variables node) ; set of variables bound by a procedure
  (list->set (proc-parms node)))

(define (not-mutable? var)
  (set-empty? (var-sets var)))

(define (mutable? var)
  (not (not-mutable? var)))

(define (bound? var)
  (var-bound var))

(define (global? var)
  (not (bound? var)))

;------------------------------------------------------------------------------

; (expression->parse-tree expr env) returns the parse tree for the Scheme
; expression 'expr' in the environment 'env'.  If 'expr' is not syntactically
; correct, an error message is output.

(define (expression->parse-tree expr env)
  (pt expr env))

(define (pt expr env)
  (cond ((unspecified-expr? expr)      (pt-unspecified expr env))
        ((self-eval-expr? expr)        (pt-self-eval expr env))
        ((quote-expr? expr)            (pt-quote expr env))
        ((quasiquote-expr? expr)       (pt-quasiquote expr env))
        ((unquote-expr? expr)          (error "Ill-placed 'unquote'"))
        ((unquote-splicing-expr? expr) (error "Ill-placed 'unquote-splicing'"))
        ((var-expr? expr)              (pt-var expr env))
        ((set!-expr? expr)             (pt-set! expr env))
        ((lambda-expr? expr)           (pt-lambda expr env))
        ((if-expr? expr)               (pt-if expr env))
        ((cond-expr? expr)             (pt-cond expr env))
        ((and-expr? expr)              (pt-and expr env))
        ((or-expr? expr)               (pt-or expr env))
        ((case-expr? expr)             (pt-case expr env))
        ((let-expr? expr)              (pt-let expr env))
        ((let*-expr? expr)             (pt-let* expr env))
        ((letrec-expr? expr)           (pt-letrec expr env))
        ((begin-expr? expr)            (pt-begin expr env))
        ((do-expr? expr)               (pt-do expr env))
        ((define-expr? expr)           (error "Ill-placed 'define'"))
        ((delay-expr? expr)            (pt-delay expr env))
        ((combination-expr? expr)      (pt-combination expr env))
        (else                          (error "Unknown expression type"))))

(define (pt-unspecified val env) (new-uns))
(define (pt-self-eval val env)   (new-cst val))
(define (pt-quote expr env)      (new-cst (second expr)))

(define (pt-quasiquote expr env)
  (pt-quasiquotation (second expr) 1 env))

(define (pt-quasiquotation form level env)
  (cond ((= level 0)
         (pt form env))
        ((quasiquote-expr? form)
         (pt-quasiquotation-list form (+ level 1) env))
        ((unquote-expr? form)
         (if (= level 1)
           (pt (second form) env)
           (pt-quasiquotation-list form (- level 1) env)))
        ((unquote-splicing-expr? form)
         (if (= level 1)
           (error "Ill-placed 'unquote-splicing'")
           (pt-quasiquotation-list form (- level 1) env)))
        ((pair? form)
         (pt-quasiquotation-list form level env))
        ((vector? form)
         (vector-form (pt-quasiquotation-list (vector->list form) level env)))
        (else
         (new-cst form))))

(define (pt-quasiquotation-list l level env)
  (if (pair? l)
    (if (and (unquote-splicing-expr? (first l)) (= level 1))
      (append-form (pt (second (first l)) env)
                   (pt-quasiquotation (after-first l) 1 env))
      (cons-form (pt-quasiquotation (first l) level env)
                 (pt-quasiquotation (after-first l) level env)))
    (pt-quasiquotation l level env)))

(define (append-form ptree1 ptree2)
  (cond ((and (cst? ptree1) (cst? ptree2))
         (new-cst (append (cst-val ptree1) (cst-val ptree2))))
        ((and (cst? ptree2) (null? (cst-val ptree2)))
         ptree1)
        (else
         (new-call
           (new-cst (primitive-procedure 'append))
           (list ptree1 ptree2)))))

(define (cons-form ptree1 ptree2)
  (cond ((and (cst? ptree1) (cst? ptree2))
         (new-cst (cons (cst-val ptree1) (cst-val ptree2))))
        ((and (cst? ptree2) (null? (cst-val ptree2)))
         (new-call (new-cst (primitive-procedure 'list)) (list ptree1)))
        (else
         (new-call
           (new-cst (primitive-procedure 'cons))
           (list ptree1 ptree2)))))

(define (vector-form ptree)
  (if (cst? ptree)
    (new-cst (list->vector (cst-val ptree)))
    (new-call (new-cst (primitive-procedure 'list->vector)) (list ptree))))

(define (pt-var name env)
  (new-ref (lookup-var name env)))

(define (pt-set! expr env)
  (new-set (lookup-var (second expr) env) (pt (third expr) env)))

(define (pt-lambda expr env)
  (let ((parms (new-parameters (second expr))))
    (new-proc (rest-param? (second expr))
              parms
              (pt-body (after-second expr) (new-env parms env)))))

(define (pt-body body env)
  (define (letrec-defines body bindings)
    (cond ((null? body)
           (error "Body must contain at least one evaluable expression"))
          ((define-expr? (first body))
           (let ((var (definition-variable (first body))))
             (if (assq var bindings)
               (error "Duplicate definition of the same variable")
               (letrec-defines (after-first body)
                               (cons (list var (definition-value (first body)))
                                     bindings)))))
          ((null? bindings)
           (cons 'begin body))
          (else
           (cons 'letrec (cons (reverse bindings) body)))))
  (pt (letrec-defines body '()) env))

(define (pt-if expr env)
  (new-cond (pt (second expr) env)
            (pt (third expr) env)
            (if (length? expr 3) (new-uns) (pt (fourth expr) env))))

(define (pt-cond expr env)
  (define (pt-clauses clauses)
    (if (length? clauses 0)
      (new-uns)
      (let ((clause (first clauses)))
        (cond ((eq? (first clause) 'else)
               (pt-sequence (after-first clause) env))
              ((length? clause 1)
               (new-disj (pt (first clause) env)
                         (pt-clauses (after-first clauses))))
              ((eq? (second clause) '=>)
               (new-disj-call (pt (first clause) env)
                              (pt (third clause) env)
                              (pt-clauses (after-first clauses))))
              (else
               (new-cond (pt (first clause) env)
                         (pt-sequence (after-first clause) env)
                         (pt-clauses (after-first clauses))))))))
  (pt-clauses (after-first expr)))

(define (pt-and expr env)
  (define (pt-exprs exprs)
    (cond ((length? exprs 0)
           (pt #t env))
          ((length? exprs 1)
           (pt (first exprs) env))
          (else
           (new-conj (pt (first exprs) env) (pt-exprs (after-first exprs))))))
  (pt-exprs (after-first expr)))

(define (pt-or expr env)
  (define (pt-exprs exprs)
    (cond ((length? exprs 0)
           (pt #f env))
          ((length? exprs 1)
           (pt (first exprs) env))
          (else
           (new-disj (pt (first exprs) env) (pt-exprs (after-first exprs))))))
  (pt-exprs (after-first expr)))

(define (pt-case expr env)
  (let ((temp (new-variables '(temp))))
    (define (pt-clauses clauses)
      (if (length? clauses 0)
        (new-uns)
        (let ((clause (first clauses)))
          (if (eq? (first clause) 'else)
            (pt-sequence (after-first clause) env)
            (new-cond (new-call (new-cst (primitive-procedure 'memv))
                                (list (new-ref (first temp))
                                      (new-cst (first clause))))
                      (pt-sequence (after-first clause) env)
                      (pt-clauses (after-first clauses)))))))
    (new-call
      (new-proc #f temp (pt-clauses (after-second expr)))
      (list (pt (second expr) env)))))

(define (pt-let expr env)
  (if (var-expr? (second expr))
    (let* ((self (new-variables (list (second expr))))
           (vars (new-variables (map first (third expr))))
           (vals (map (lambda (x) (pt (second x) env)) (third expr)))
           (env  (new-env self (new-env vars env))))
      (new-call
        (new-proc #f self (new-call (new-ref (first self)) vals))
        (list (new-proc #f vars (pt-body (after-third expr) env)))))
    (if (null? (second expr))
      (pt-body (after-second expr) env)
      (let* ((vars (new-variables (map first (second expr))))
             (vals (map (lambda (x) (pt (second x) env)) (second expr)))
             (env  (new-env vars env)))
        (new-call (new-proc #f vars (pt-body (after-second expr) env)) vals)))))

(define (pt-let* expr env)
  (if (null? (second expr))
    (pt-body (after-second expr) env)
    (pt (list 'let (list (first (second expr)))
          (cons 'let* (cons (after-first (second expr)) (after-second expr))))
        env)))

(define (pt-letrec expr env)
  (let* ((vars  (new-variables (map first (second expr))))
         (env   (new-env vars env))
         (vals  (map (lambda (x) (pt (second x) env)) (second expr))))
    (define (val-of var)
      (list-ref vals (- (length vars) (length (memq var vars)))))
    (define (bind-in-order order)
      (if (null? order)
        (pt-body (after-second expr) env)
        (let ((vars (set->list (first order))))
          (new-call
            (new-proc #f vars (bind-in-order (after-first order)))
            (map val-of vars)))))
    (bind-in-order (topological-sort
                     (transitive-closure
                       (dependency-graph vars vals))))))

(define (pt-begin expr env)
  (pt-sequence (after-first expr) env))

(define (pt-sequence expr-list env)
  (if (length? expr-list 1)
    (pt (first expr-list) env)
    (new-seq (pt (first expr-list) env)
             (pt-sequence (after-first expr-list) env))))

(define (pt-do expr env)
  (let* ((loop (new-variables '(loop)))
         (vars (new-variables (map first (second expr))))
         (init (map (lambda (x) (pt (second x) env)) (second expr)))
         (env  (new-env vars env))
         (step (map (lambda (x)
                      (pt (if (length? x 2) (first x) (third x)) env))
                    (second expr))))
    (new-call
      (new-proc #f loop (new-call (new-ref (first loop)) init))
      (list
        (new-proc #f vars
          (new-cond (pt (first (third expr)) env)
                    (if (length? (third expr) 1)
                      (new-uns)
                      (pt-sequence (after-first (third expr)) env))
                    (if (length? expr 3)
                      (new-call (new-ref (first loop)) step)
                      (new-seq (pt-sequence (after-third expr) env)
                               (new-call (new-ref (first loop)) step)))))))))

(define (pt-delay expr env)
  (new-call
    (new-cst (primitive-procedure 'make-delayed))
    (list (new-proc #f '() (pt (second expr) env)))))

(define (pt-combination expr env)
  (new-call (pt (first expr) env)
            (map (lambda (x) (pt x env)) (after-first expr))))

; Expression identification predicates and syntax checking.

(define (unspecified-expr? x) (eq? x unspecified))
(define unspecified unspecified-expr?)

(define (self-eval-expr? expr)
  (and (not (pair? expr)) (not (symbol? expr))))

(define (quote-expr? expr)
  (match 'quote 1 expr))

(define (quasiquote-expr? expr)
  (match 'quasiquote 1 expr))

(define (unquote-expr? expr)
  (match 'unquote 1 expr))

(define (unquote-splicing-expr? expr)
  (match 'unquote-splicing 1 expr))

(define (var-expr? expr)
  (if (syntactic-keyword? expr)
    (error "Variable name can not be a syntactic keyword")
    (symbol? expr)))

(define syntactic-keywords
  '(quote quasiquote unquote unquote-splicing lambda if set! cond => else
    and or case let let* letrec begin do define delay))

(define (syntactic-keyword? expr)
  (memq expr syntactic-keywords))

(define (set!-expr? expr)
  (and (match 'set! 2 expr) (var-expr? (second expr))))

(define (lambda-expr? expr)
  (and (match 'lambda -2 expr) (proper-parms? (second expr))))

(define (if-expr? expr)
  (and (match 'if -2 expr)
       (or (<= (length expr) 4) (error "Ill-formed special form" 'if))))

(define (cond-expr? expr)
  (and (match 'cond -1 expr) (proper-clauses? (after-first expr))))

(define (and-expr? expr)
  (match 'and 0 expr))

(define (or-expr? expr)
  (match 'or 0 expr))

(define (case-expr? expr)
  (and (match 'case -2 expr) (proper-case-clauses? (after-second expr))))

(define (let-expr? expr)
  (and (match 'let -2 expr)
       (if (var-expr? (second expr))
         (and (proper-bindings? (third expr))
              (or (> (length expr) 3) (error "Ill-formed named let")))
         (proper-bindings? (second expr)))))

(define (let*-expr? expr)
  (and (match 'let* -2 expr) (proper-bindings? (second expr))))

(define (letrec-expr? expr)
  (and (match 'letrec -2 expr) (proper-bindings? (second expr))))

(define (begin-expr? expr)
  (match 'begin -1 expr))

(define (do-expr? expr)
  (and (match 'do -2 expr)
       (proper-do-bindings? (second expr))
       (proper-do-exit? (third expr))))

(define (define-expr? expr)
  (and (match 'define -1 expr)
       (proper-definition? (second expr) (after-second expr))))

(define (delay-expr? expr)
  (match 'delay 1 expr))

(define (combination-expr? expr)
  (let ((length (proper-length expr)))
    (if length
      (or (> length 0) (error "Ill-formed procedure call"))
      (error "Ill-terminated procedure call"))))

(define (match keyword size expr)
  (and (pair? expr) (eq? (first expr) keyword)
       (let ((length (proper-length (after-first expr))))
         (if length
           (or (if (> size 0) (= length size) (>= length (- size)))
               (error "Ill-formed special form" keyword))
           (error "Ill-terminated special form" keyword)))))

(define (proper-length list)
  (define (length list n)
    (cond ((null? list) n)
          ((pair? list) (length (after-first list) (+ n 1)))
          (else         #f)))
  (length list 0))

(define (proper-definition? pattern body)
  (cond ((var-expr? pattern)
         (cond ((length? body 0) #t) ; an unbound variable
               ((length? body 1) #t) ; a bound variable
               (else             (error "Ill-formed definition body"))))
        ((pair? pattern)
         (if (length? body 0)
           (error "Body of a definition must have at least one expression"))
         (if (var-expr? (first pattern))
           (proper-parms? (after-first pattern))
           (error "Procedure name must be an identifier")))
        (else
         (error "Ill-formed definition pattern"))))

(define (definition-variable def)
  (if (pair? (second def))
    (first (second def))
    (second def)))

(define (definition-value def)
  (cond ((pair? (second def))
         (cons 'lambda (cons (after-first (second def)) (after-second def))))
        ((null? (after-second def))
         unspecified)
        (else
         (third def))))

(define (proper-parms? parms)
  (define (proper-parms parms seen)
    (cond ((pair? parms)
           (if (var-expr? (first parms))
             (if (memq (first parms) seen)
               (error "Duplicate parameter in parameter list")
               (proper-parms (after-first parms) (cons (first parms) seen)))
             (error "Parameter must be an identifier")))
          ((null? parms)
           #t)
          ((var-expr? parms)
           (if (memq parms seen)
             (error "Duplicate parameter in parameter list")
             #t))
          (else
           (error "Rest parameter must be an identifier"))))
  (proper-parms parms '()))

(define (proper-clauses? clauses)
  (or (null? clauses)
      (let ((length (proper-length (first clauses))))
        (if length
          (if (>= length 1)
            (if (eq? (first (first clauses)) 'else)
              (cond ((= length 1)
                     (error "Else clause must have a body"))
                    ((not (null? (after-first clauses)))
                     (error "Else clause must be the last clause"))
                    (else
                     (proper-clauses? (after-first clauses))))
              (if (and (>= length 2)
                       (eq? (second (first clauses)) '=>)
                       (not (= length 3)))
                (error "'=>' must be followed by a single expression")
                (proper-clauses? (after-first clauses))))
            (error "Ill-formed 'cond' clause"))
          (error "Ill-terminated 'cond' clause")))))

(define (proper-case-clauses? clauses)
  (or (null? clauses)
      (let ((length (proper-length (first clauses))))
        (if length
          (if (>= length 2)
            (if (eq? (first (first clauses)) 'else)
              (or (null? (after-first clauses))
                  (error "Else clause must be the last clause"))
              (proper-selector-list? (first (first clauses))))
            (error "A 'case' clause must have a selector list and a body"))
          (error "Ill-terminated 'case' clause")))))

(define (proper-selector-list? l)
  (let ((length (proper-length l)))
    (if length
      (or (>= length 1)
          (error "Selector list must contain at least one element"))
      (error "Ill-terminated selector list"))))

(define (proper-bindings? bindings)
  (define (proper-bindings bindings seen)
    (cond ((null? bindings)
           #t)
          ((not (pair? bindings))
           (error "Ill-terminated binding list"))
          (else
           (let* ((binding (first bindings))
                  (length (proper-length binding)))
             (if (eqv? length 2)
               (if (var-expr? (first binding))
                 (if (memq (first binding) seen)
                   (error "Duplicate variable in bindings")
                   (proper-bindings (after-first bindings)
                                    (cons (first binding) seen)))
                 (error "Binding variable must be an identifier"))
               (error "Ill-formed binding"))))))
   (proper-bindings bindings '()))

(define (proper-do-bindings? bindings)
  (define (proper-bindings bindings seen)
    (cond ((null? bindings)
           #t)
          ((not (pair? bindings))
           (error "Ill-terminated binding list"))
          (else
           (let* ((binding (first bindings))
                  (length (proper-length binding)))
             (if (or (eqv? length 2) (eqv? length 3))
               (if (var-expr? (first binding))
                 (if (memq (first binding) seen)
                   (error "Duplicate variable in bindings")
                   (proper-bindings (after-first bindings)
                                    (cons (first binding) seen)))
                 (error "Binding variable must be an identifier"))
               (error "Ill-formed binding"))))))
   (proper-bindings bindings '()))

(define (proper-do-exit? exit)
  (let ((length (proper-length exit)))
    (if length
      (or (> length 0) (error "Ill-formed exit clause"))
      (error "Ill-terminated exit clause"))))

; Utility procedures.

(define (length? l n)
  (cond ((null? l) (= n 0))
        ((> n 0)   (length? (after-first l) (- n 1)))
        (else      #f)))

(define (first l)  (car l))
(define (second l) (cadr l))
(define (third l)  (caddr l))
(define (fourth l) (cadddr l))

(define (after-first l)  (cdr l))
(define (after-second l) (cddr l))
(define (after-third l)  (cdddr l))

;------------------------------------------------------------------------------

; Variable dependency analysis for recursive definitions (e.g. 'letrec's).

(define-structure gnode label edges)

(define (dependency-graph vars vals)
  (define (dgraph vars* vals*)
    (if (null? vars*)
      set-empty
      (let ((var (first vars*)) (val (first vals*)))
        (set-adjoin (dgraph (after-first vars*) (after-first vals*))
                    (make-gnode var (set-intersection
                                      (list->set vars)
                                      (free-variables val)))))))
  (dgraph vars vals))

(define (transitive-closure graph)
  (define changed? #f)
  (define (closure edges)
    (list->set (apply set-union edges
                 (map (lambda (label) (gnode-edges (gnode-find label graph)))
                      (set->list edges)))))
  (let ((new-graph
          (set-map (lambda (x)
                     (let ((new-edges (closure (gnode-edges x))))
                       (if (not (set-equal? new-edges (gnode-edges x)))
                         (set! changed? #t))
                       (make-gnode (gnode-label x) new-edges)))
                   graph)))
    (if changed? (transitive-closure new-graph) new-graph)))

(define (gnode-find label graph)
  (define (find label l)
    (cond ((null? l)                           #f)
          ((eq? (gnode-label (first l)) label) (first l))
          (else                                (find label (after-first l)))))
  (find label (set->list graph)))

(define (topological-sort graph) ; topological sort fixed to handle cycles
  (if (set-empty? graph)
    '()
    (let ((to-remove (or (remove-no-edges graph) (remove-cycle graph))))
      (let ((labels (set-map gnode-label to-remove)))
        (cons labels
              (topological-sort
                (set-map (lambda (x)
                           (make-gnode
                             (gnode-label x)
                             (set-difference (gnode-edges x) labels)))
                         (set-difference graph to-remove))))))))

(define (remove-no-edges graph)
  (let ((nodes-with-no-edges
         (set-keep (lambda (x) (set-empty? (gnode-edges x))) graph)))
    (if (set-empty? nodes-with-no-edges)
      #f
      nodes-with-no-edges)))

(define (remove-cycle graph)
  (define (remove l)
    (let ((edges (gnode-edges (first l))))
      (define (equal-edges? x) (set-equal? (gnode-edges x) edges))
      (define (member-edges? x) (set-member? (gnode-label x) edges))
      (if (set-member? (gnode-label (first l)) edges)
        (let ((edge-graph (set-keep member-edges? graph)))
          (if (set-every? equal-edges? edge-graph)
            edge-graph
            (remove (after-first l))))
        (remove (after-first l)))))
  (remove (set->list graph)))

;------------------------------------------------------------------------------

; (parse-tree->expression ptree) returns the Scheme expression corresponding to
; the parse tree 'ptree'.

(define (parse-tree->expression ptree)
  (se ptree '() (list 0)))

(define (se ptree env num)
  (cond ((root? ptree)
         (error "Root nodes should not appear inside parse trees"))
        ((uns? ptree)
         ''<unspecified>)
        ((cst? ptree)
         (list 'quote (cst-val ptree)))
        ((ref? ptree)
         (let ((x (assq (ref-var ptree) env)))
           (if x (cdr x) (var-name (ref-var ptree)))))
        ((set? ptree)
         (list 'set!
           (let ((x (assq (set-var ptree) env)))
             (if x (cdr x) (var-name (set-var ptree))))
           (se (set-val ptree) env num)))
        ((cond? ptree)
         (list 'if (se (cond-pre ptree) env num)
                   (se (cond-con ptree) env num)
                   (se (cond-alt ptree) env num)))
        ((proc? ptree)
         (let ((new-env (se-rename (proc-parms ptree) env num)))
           (list 'lambda
             (se-parameters (proc-parms ptree) (proc-rest ptree) new-env)
             (se (proc-body ptree) new-env num))))
        ((call? ptree)
         (let ((oper (call-oper ptree))
               (args (call-args ptree)))
           (if (and (proc? oper) ; applying a lambda-expr is like a 'let'
                    (not (proc-rest oper))
                    (= (length (proc-parms oper)) (length args)))
             (let ((new-env (se-rename (proc-parms oper) env num)))
               (list
                 (if (set-empty?
                       (set-intersection
                         (list->set (proc-parms oper))
                         (apply set-union (map free-variables args))))
                   'let
                   'letrec)
                 (se-bindings (proc-parms oper) args new-env num)
                 (se (proc-body oper) new-env num)))
             (map (lambda (x) (se x env num)) (cons oper args)))))
        (else
         (error "Unknown parse tree node type"))))

(define (se-parameters parms rest env)
  (cond ((null? parms)
         '())
        ((and rest (null? (after-first parms)))
         (cdr (assq (first parms) env)))
        (else
         (cons (cdr (assq (first parms) env))
               (se-parameters (after-first parms) rest env)))))

(define (se-bindings vars vals env num)
  (if (null? vars)
    '()
    (cons (list (cdr (assq (first vars) env)) (se (first vals) env num))
          (se-bindings (after-first vars) (after-first vals) env num))))

(define (se-rename vars env num)
  (define (rename vars)
    (if (null? vars)
      env
      (cons (cons (first vars)
                  (string->symbol
                    (string-append (symbol->string (var-name (first vars)))
                                   "_"
                                   (number->string (car num)))))
            (rename (after-first vars)))))
  (set-car! num (+ (car num) 1))
  (rename vars))

;------------------------------------------------------------------------------
