;------------------------------------------------------------------------------
;
; STRUCTURE definition package
;
; This package implements `structure' types for Scheme.  To define a structure
; type, the macro `define-structure' should be used, specifying the name of the
; structure and each of its fields.  This will declare some functions to:
; create instances of this type, determine if an object is of the given
; structure type and select and mutate fields of such objects.  A new `define'
; macro is also declared to define sub-classes of the structure (i.e. a
; structure which has the fields of its super-class but with some more fields).
; Here is a sample use of `define-structure':
;
; (define-structure person name sex age)
; (define-person student field)
; (define-person priest religion)
; (define x (make-person "Jack" 'male 21))
; (define y (make-person "Jill" 'female)) ; we have omitted the age...
; (define z (make-student "Bob" 'male 25 'biology))
; (person? (cons 1 2)) ==> #f
; (person? x) ==> #t
; (person? z) ==> #t
; (student? x) ==> #f
; (student? z) ==> #t
; (person-name x) ==> "Jack"
; (person-age y) ==> ()
; (person-age-set! x 22) ; change the age of "Jack" to 22
; (person-name z) ==> "Bob"
; (student-field z) ==> biology
;
;------------------------------------------------------------------------------

(define *global-environment* (the-environment)) ; where to put the functions

(define (define-substructure add-substructure super-fields name fields)
  (let* ((all-fields (append super-fields fields))
         (tag (cons name all-fields))
         (tags (list tag))
         (struct-length (+ (length all-fields) 1))
         (s-name (symbol->string name)))
    (define (symbol . strings)
      (string->symbol (apply string-append strings)))
    (define (gdefine name value)
      (list 'define name (list 'quote value)))
    (define (is-a-struct? x)
      (and (vector? x) (> (vector-length x) 0) (memq (vector-ref x 0) tags)))
    (define (make-struct . values)
      (let ((struct (make-vector struct-length)))
        (vector-set! struct 0 tag)
        (do ((v values (cdr v)) (f all-fields (cdr f)) (i 1 (+ i 1)))
            ((or (null? v) (null? f))
             (if (not (null? v)) (error "too many values for a structure")))
          (vector-set! struct i (car v)))
        struct))
    (define (struct-selector field)
      (let ((pos (- struct-length (length (memq field fields)))))
        (lambda (x) (if (is-a-struct? x)
                      (vector-ref x pos) 
                      (error "invalid argument to a structure selector")))))
    (define (struct-mutator field)
      (let ((pos (- struct-length (length (memq field fields)))))
        (lambda (x y) (if (is-a-struct? x)
                        (vector-set! x pos y) 
                        (error "invalid argument to a structure mutator")))))
    (add-substructure tag)
    (list 'eval
      (append
        (list
          'begin
          (list 'define-macro
            (cons (symbol "DEFINE-" s-name) '(name . fields))
            (list (list 'quote define-substructure)
              (list 'quote (lambda (x) (set! tags (cons x tags))))
              (list 'quote all-fields)
              'name
              'fields))
          (gdefine (symbol "MAKE-" s-name) make-struct)
          (gdefine (symbol s-name "?") is-a-struct?))
        (map
          (lambda (field)
            (gdefine (symbol s-name "-" (symbol->string field))
              (struct-selector field)))
          fields)
        (map
          (lambda (field)
            (gdefine (symbol s-name "-" (symbol->string field) "-SET!")
              (struct-mutator field)))
          fields))
      '*global-environment*)))

(eval
  (list 'define-macro '(define-structure name . fields)
    (list (list 'quote define-substructure)
          (list 'quote (lambda (x) x))
          '()
          'name
          'fields))
  *global-environment*)

;------------------------------------------------------------------------------
