/* SIS Scheme UNIX interface for SUN-3's */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/times.h>

#define K 1024

#define null -3
#define undefined -7

#define vector_type      0x01
#define symbol_type      0x02
#define stack_type       0x20
#define string_type	 0x30
#define procedure_type   0x8000
#define primitive_type   0xa000
#define closure_type     0xc000
#define entry_label_type 0xe000
#define jump_opcode      0x4ef9
#define call_opcode      0x4eb9
#define trap_opcode      0x4e95

typedef char *PTR;

extern void sis_entry();
extern char process_transfer;
extern char pair_area1;
extern char pair_area2;

/*---------------------------------------------------------------------------*/

/* error message output */

void warning( msg )
char *msg;
{
  fprintf( stderr, "sis: warning, %s\n", msg );
}

void fatal( msg )
char *msg;
{
  fprintf( stderr, "sis: %s\n", msg );
  exit( 1 );
}

/*---------------------------------------------------------------------------*/

long char_buffer = -1;

long read_char()
{
  long temp;

  if (char_buffer < 0) char_buffer = getchar();
  temp = char_buffer;
  char_buffer = -1;
  return temp;

}

long peek_char()
{

  if (char_buffer < 0) char_buffer = getchar();
  return char_buffer;

}

void write_char( c )
long c;
{

  putchar( c );

}

struct tms buffer;

long cpu_time()
{
  times(&buffer);
  return buffer.tms_utime;

}

/*---------------------------------------------------------------------------*/

/* memory allocation of blocks starting at an octuple address */

PTR malloc8( length )
long length;
{
  PTR ptr;

  ptr = (PTR) malloc( length+7 );
  if (ptr != NULL) ptr = (PTR) (((long) ptr + 7) & -8); /* get octuple addr. */
  return ptr;
}

/*---------------------------------------------------------------------------*/

/* constants and variables in library table */

extern char library_table; /* middle of library table */

extern long fix_mask;   /* the mask to test for fixnums (contained in D5) */
extern long pair_mask;  /* the mask to test for pairs (contained in D6)   */
extern long hunk_mask;  /* the mask to test for hunks (contained in D7)   */
extern long empty_list; /* the empty list object (contained in A4)        */
extern PTR libr_ptr;    /* pointer to the library table (contained in A5) */
extern PTR glob_ptr;    /* pointer to the global table (contained in A6)  */

extern PTR heap_ptr;   /* pointer to start of heap                        */
extern PTR sub_heap_ptr;
extern long sub_heap_size;
extern PTR heap_alloc_ptr;
extern long free_count;

extern PTR current_process;

void init_library_table()
{

/* the library table is a block of memory organized as follows:

               _________
              |    .    | \
              |    .    |  | system variables & constants
              |_________| /
              |    .    | \
libr_ptr      |    .    |  | primitive procedures (car, cdr, +, -, etc.)
(reg A5)      |_________| /
----------> 0 |    .    | \
              |    .    |  | trap handler code
              |_________| /
              |    .    | \
              |    .    |  | other primitive procedures
              |_________| /

The primitive procedures are spread around libr_ptr so that the 68000 can
address them using a 16 bit offset.  The trap handler is called with a
'jsr a5@' instruction.  This table is part of the program's image so it does
not have to be initialized explicitly (see the file 'library.s' for the code
contained in this table).

*/

  if ((((long) &library_table) & 7) != 0)
    fatal( "internal error, library table is not at an octuple address" );

  libr_ptr = &library_table;

}

/*---------------------------------------------------------------------------*/

#define glob_var_count (* (short *) (glob_ptr-2))
#define stack_limit_ptr (* (PTR *) glob_ptr)
#define GLOBAL_VAR_OPCODE(i) (* (short *) (glob_ptr+(i*6+33022)%65536-32768))
#define GLOBAL_VAR(i) (* (long *) (glob_ptr+(i*6+33024)%65536-32768))

void init_global_table()
{

/* the global table is a 64K byte block of memory organized as follows:

              <-16 bits->
               _________
       -32768 |    .    | \
              |    .    |  | slots for last 5461 global variables
              |    .    |  |
glob_ptr      |_________| /
(reg A6)   -2 |_________| global variable count
----------> 0 |_________| \ stack limit pointer
              |_________| /
            4 |    .    | \
              |    .    |  | code for global var. assignment and cache clear
              |    .    |  | (32 *  movl d1,a1@ / movw #trap_opcode,a1@- / rts)
              |_________| /
          260 |    .    | \
              |    .    |  | slots for first 5418 global variables (each var.
              |    .    |  | occupies 6 bytes, the first 2 are for an opcode
              |_________| /  the following 4 are for the variable's value)

*/

  PTR temp = malloc8( 64*K );

  if (temp == NULL)
    fatal( "insufficient storage to allocate global table" );
  else {

    glob_ptr = temp + 0x8000; /* adjust pointer to middle of block */

    glob_var_count  = 0; /* no global variables yet */
    stack_limit_ptr = 0; /* no stack limit yet      */

    { int i; short *p = (short *) (glob_ptr+4); /* init assignment handlers */
      for( i=1; i<=32; i++ ) {
        *p++ = 0x2281;      /* movl d1,a1@            */
        *p++ = 0x333c;      /* movw #trap_opcode,a1@- */
        *p++ = trap_opcode; /* opcode for: jsr a5@    */
        *p++ = 0x4e75;      /* rts                    */
      }
    }

    { int i;                                 /* init global variables     */
      for( i=1; i<=10879; i++ ) {
        GLOBAL_VAR_OPCODE(i) = trap_opcode;  /* jsr a5@                   */
        GLOBAL_VAR(i)        = undefined;    /* initial value = undefined */
      }
    }

  }

}

/*---------------------------------------------------------------------------*/

void init_heap( length )
long length;
{

/* the heap is a block of memory containing two equaly sized sub-heaps
   each starting on an octuple address:

heap_ptr       _________
------------> |    .    | \
              |    .    |  | sub-heap 1 (first to be used)
              |    .    |  |
              |_________| /
              |    .    | \
              |    .    |  | sub-heap 2
              |    .    |  |
              |_________| /

*/

  if (length < 64) {
    length = 64;
    warning( "minimum size heap (64K) is being allocated" );
  }

  heap_ptr = malloc8( length*K );

  if (heap_ptr == NULL)
    fatal( "insufficient storage to allocate heap" );
  else {
    sub_heap_ptr   = heap_ptr;
    sub_heap_size  = length*K/2;
    heap_alloc_ptr = sub_heap_ptr;
    free_count     = sub_heap_size/4  *2 /*use both halves since no GC*/;
  }

}

/*---------------------------------------------------------------------------*/

void main_err()
{
  fatal( "Usage: sis [-h heap_size_in_kilobytes] [-f state_file_name]" );
}

void main( argc, argv )
int argc; char *argv[];
{
  long heap_length;
  char *state_file, *arg;
  int i;
  PTR current_stack;

  heap_length = 2048;
  state_file  = "sis.state";

  for( i=1; i<argc; i++ ) {
    arg = argv[i];
    if (*arg == '-') {
      if (*++arg == 'h') heap_length = atol( ++arg );
      else if (*arg == 'f') state_file = ++arg;
      else
        main_err();
      }
    else
      main_err();
    }

  if ((((long) &pair_area1) & 7) != 0)
    fatal( "internal error, pair area 1 is not at an octuple address" );

  if ((((long) &pair_area2) & 7) != 0)
    fatal( "internal error, pair area 2 is not at an octuple address" );

  init_library_table();

  init_global_table();

  init_heap( heap_length );

  fix_mask   = 0x01010101;
  pair_mask  = 0x10101010;
  hunk_mask  = 0x44444444;
  empty_list = null;

  current_stack                   = heap_alloc_ptr+2;
  (* (long *) (current_stack-2))  = (stack_type<<24) + 10000*4;
  heap_alloc_ptr                  += 10001*4;
  free_count                      -= 10001;
  (* (long *) (heap_alloc_ptr-4)) = 0;

  current_process                  = heap_alloc_ptr+2;
  (* (long *) (current_process-2)) = ((closure_type+2)<<16) + jump_opcode;
  heap_alloc_ptr                   += 3*4;
  free_count                       -= 3;
  (* (PTR *) (current_process+2))  = &process_transfer;
  (* (PTR *) (current_process+6))  = current_stack;

  sis_entry();

}
