;; -----------------------------------------------------------------------
;;                          CS110 - Assignment #2
;;                 Implementing an associative-like memory
;;
;;                      by: Marc Feeley (oct 1, 1987)
;; -----------------------------------------------------------------------


;; ======================= Pseudo-random generator =======================

;;; (random-int) returns a random integer in the range
;;; 0 <= x < *random-limit* (uses linear congruential method).

(define (random-int)
  (set! *seed* (remainder (+ (* *seed* 3581) 12751) *random-limit*))
  *seed*)

(define *seed* 0) ; initial value of seed

(define *random-limit* 32768) ; 2**15


;; ========================== Memory management ==========================


;;; (make-memory nb-loc nb-bits dist) returns an associative-like memory
;;; object containing 'nb-loc' locations and 'nb-bits' bit addresses.
;;; Locations at a hamming distance 'dist' of a given address will be
;;; examined by the read/write operations.

(define (make-memory nb-loc nb-bits dist)

  ;;; 'memory' is a vector of pairs; the car of these pairs is an address
  ;;; and the cdr the contents at that address.  All addresses are unique.

  (define memory (make-vector nb-loc))

  ;;; (make-address addr) returns the binary representation of the
  ;;; numerical address 'addr'.  (i.e. it converts an integer into a
  ;;; vector of 0's and 1's.

  (define (make-address addr)
    (let ((address (make-vector nb-bits)))
      (do ((i (- nb-bits 1) (- i 1)) (n addr (quotient n 2)))
          ((< i 0) address)
        (vector-set! address i (remainder n 2)))))

  ;;; (init-addresses n) assigns unique addresses to the memory cells.
  ;;; The addresses are chosen randomly from the range 0 to 'n'-1.  It is
  ;;; based on a selection sampling technique from Knuth vol. 2, p. 137.
  ;;; It works best when nb-loc and 2**nb-bits are of comparable size.

  (define (init-addresses n)
    (do ((i 0 (+ i 1)) (s 0)) ((= s nb-loc))
      (if (< (* (- n i) (random-int)) (* (- nb-loc s) *random-limit*))
        (begin
          (vector-set! memory s (cons (make-address i)
                                      (make-vector nb-bits 0)))
          (set! s (+ s 1))))))

  ;;; (distance address1 address2) gives the hamming distance between the
  ;;; two addresses (i.e. the number of bit positions at which the
  ;;; addresses differ).

  (define (distance address1 address2)
    (do ((i (- nb-bits 1) (- i 1)) (d 0)) ((< i 0) d)
      (if (not (= (vector-ref address1 i) (vector-ref address2 i)))
        (set! d (+ d 1)))))

  ;;; (write address . word) stores a word in memory at a given address.
  ;;; If the word is not specified, the address is stored.

  (define (write address . word)
    (define (store addr val)
      (do ((i 0 (+ i 1))) ((= i nb-loc)) ; scan all locations
        (if (<= (distance addr (car (vector-ref memory i))) dist)
          (let ((counters (cdr (vector-ref memory i))))
            (do ((j (- nb-bits 1) (- j 1))) ((< j 0))
              (vector-set! counters j
                (+ (vector-ref counters j)
                   (if (= (vector-ref val j) 1) 1 -1))))))))
    (store address (if (null? word) address (car word))))

  ;;; (read address . iterations) retrieves a word from the memory given
  ;;; its 'address'.  The retrieval operation is repeated (using the last
  ;;; obtained value as the address) until the value retrieved is the
  ;;; same as the last address (repeated up to a maximum of 'iterations'
  ;;; times, default is 1).

  (define (read address . iterations)
    (define (retrieve addr)
      (do ((i 0 (+ i 1)) (sums (make-vector nb-bits 0)))
          ((= i nb-loc) ; scan all locations
           (do ((j 0 (+ j 1)) (result (make-vector nb-bits)))
               ((= j nb-bits) result)
             (vector-set! result j (if (< (vector-ref sums j) 0) 0 1))))
        (if (<= (distance address (car (vector-ref memory i))) dist)
          (let ((counters (cdr (vector-ref memory i))))
            (do ((j 0 (+ j 1))) ((= j nb-bits))
              (vector-set! sums j
                (+ (vector-ref sums j) (vector-ref counters j))))))))
    (do ((i (if (null? iterations) 1 (car iterations)) (- i 1))
         (a address r)
         (r (retrieve address) (retrieve r)))
        ((or (<= i 1) (equal? a r)) r)))

  (init-addresses (expt 2 nb-bits)) ; initialize memory

  (lambda (operation)
    (cond ((eq? operation 'read)  read)
          ((eq? operation 'write) write)
          (else                   (error "undefined operation")))))


;;; (memory-write memory address . word) stores 'word' in 'memory' at a
;;; given 'address'.  If the word is not specified, the address is stored.

(define (memory-write memory address . word)
  (apply (memory 'write) (cons address word)))

;;; (memory-read memory address . iterations) retrieves a word from the
;;; 'memory' given its 'address'.  The operation is repeated up to a
;;; maximum of 'iterations' times.

(define (memory-read memory address . iterations)
  (apply (memory 'read) (cons address iterations)))

;;; (memory-list memory address) returns the list of the addresses of the
;;; elements of the linked list whose head is at 'address' in 'memory'.

(define (memory-list memory address)
  (define (search-list addr contents list)
    (if (equal? addr contents) ; end of list if last element of the
      (reverse list)           ; list points to itself
      (let ((x (member contents list))) ; check if cyclic list
        (if x
          (reverse (cons contents list)) ; if cyclic, repeat address
          (search-list contents
                       (memory-read memory contents)
                       (cons contents list))))))
  (search-list address (memory-read memory address) '()))


;; ================================ Tests ================================


; First and second task of homework (Auto-association and Retrieval):

(define (test1)
  (define mem (make-memory 1000 11 3))

  (define elephant  #(1 0 1 1 0 0 0 0 1 1 0)) ; elephant NOT carnivore!
  (define shark     #(0 1 0 0 0 1 0 1 0 0 0))
  (define human     #(1 0 1 1 0 1 0 1 1 0 0))
  (define bee       #(0 0 1 1 0 0 1 0 0 0 0))
  (define robin     #(0 0 1 1 1 0 1 0 1 0 1)) ; robin has a beak!
  (define whale     #(1 1 0 0 0 1 0 1 1 1 0))
  (define alligator #(0 0 0 1 0 1 0 1 0 0 0))
  (define penguin   #(0 0 1 1 1 1 0 0 1 0 1))
  (define gorilla   #(1 0 1 1 1 0 0 1 1 0 0))

  (memory-write mem elephant )
  (memory-write mem shark    )
  (memory-write mem human    )
  (memory-write mem bee      )
  (memory-write mem robin    )
  (memory-write mem whale    )
  (memory-write mem alligator)

  (write (memory-read mem elephant )) (newline)
  (write (memory-read mem shark    )) (newline)
  (write (memory-read mem human    )) (newline)
  (write (memory-read mem bee      )) (newline)
  (write (memory-read mem robin    )) (newline)
  (write (memory-read mem whale    )) (newline)
  (write (memory-read mem alligator)) (newline)

  (write (memory-read mem penguin 10)) (newline)

  (write (memory-read mem gorilla 10)) (newline)
)

; Third task of homework (Sequence):

; To encode the data we use the following special addresses:
;
; #(0 0 0 1 0 1 1 0 1 1 0)
; #(0 0 1 0 0 1 0 1 0 1 1)
; #(0 1 0 0 0 0 1 1 1 0 1)
; #(0 1 0 1 1 0 0 1 0 1 0)
; #(1 0 0 0 1 0 0 0 1 1 1)
; #(1 1 0 0 1 1 1 0 0 0 0)
; #(1 1 1 1 0 0 0 0 0 0 1)
;
; These addresses have been chosen to be as different from each other
; and the encoding for the different animals as possible.  In fact, they
; are all at more than a distance of 3 from each other.  The interference
; between them (for read/write operations) is thus minimized.  (note:
; there are many other set of addresses with this property).

(define (test2)
  (define mem (make-memory 1000 11 3))

  (define animal    #(0 0 0 1 0 1 1 0 1 1 0))
  (define mammal    #(0 0 1 0 0 1 0 1 0 1 1))
  (define fish      #(0 1 0 0 0 0 1 1 1 0 1))
  (define bird      #(0 1 0 1 1 0 0 1 0 1 0))
  (define reptile   #(1 0 0 0 1 0 0 0 1 1 1))
  (define insect    #(1 1 0 0 1 1 1 0 0 0 0))
  (define elephant  #(1 0 1 1 0 0 0 0 1 1 0))
  (define shark     #(0 1 0 0 0 1 0 1 0 0 0))
  (define human     #(1 0 1 1 0 1 0 1 1 0 0))
  (define bee       #(0 0 1 1 0 0 1 0 0 0 0))
  (define robin     #(0 0 1 1 1 0 1 0 1 0 1))
  (define whale     #(1 1 0 0 0 1 0 1 1 1 0))
  (define alligator #(0 0 0 1 0 1 0 1 0 0 0))
  (define penguin   #(0 0 1 1 1 1 0 0 1 0 1))
  (define gorilla   #(1 0 1 1 1 0 0 1 1 0 0))

  (memory-write mem animal    animal ) ; root of tree points to itself
  (memory-write mem mammal    animal )
  (memory-write mem fish      animal )
  (memory-write mem bird      animal )
  (memory-write mem reptile   animal )
  (memory-write mem insect    animal )
  (memory-write mem elephant  mammal )
  (memory-write mem shark     fish   )
  (memory-write mem human     mammal )
  (memory-write mem bee       insect )
  (memory-write mem robin     bird   )
  (memory-write mem whale     mammal )
  (memory-write mem alligator reptile)
  (memory-write mem penguin   bird   )
  (memory-write mem gorilla   mammal )

  (write (memory-list mem elephant )) (newline)
  (write (memory-list mem shark    )) (newline)
  (write (memory-list mem human    )) (newline)
  (write (memory-list mem bee      )) (newline)
  (write (memory-list mem robin    )) (newline)
  (write (memory-list mem whale    )) (newline)
  (write (memory-list mem alligator)) (newline)
  (write (memory-list mem penguin  )) (newline)
  (write (memory-list mem gorilla  )) (newline)
)

(test1)

(newline)

(test2)
