;_______________________________________________________________________________________________
;
;                      Compilateur Scheme non-optimisant ecrit en Scheme
;
; Note: Seulement les formes speciales 'quote', 'set!', 'if' et 'lambda' ainsi que la reference
;       de constante et de variable et l'application de fonction sont traitees car les autres
;       constructions de Scheme peuvent s'exprimer a partir de celles-ci.
;_______________________________________________________________________________________________


(define (compile expr)
  ((gen expr) *env-glo*))

(define (gen expr)
  (cond ((symbol? expr)          (gen-ref expr))
        ((not (pair? expr))      (gen-cst expr))
        ((eq? (car expr) 'quote) (gen-cst (cadr expr)))
        ((eq? (car expr) 'set!)  (gen-set (cadr expr) (gen (caddr expr))))
        ((eq? (car expr) 'if)    (gen-tst (gen (cadr expr)) (gen (caddr expr)) (gen (cadddr expr))))
        ((eq? (car expr) 'lambda)
         (let ((p (cadr expr)))
           (cond ((null? p)          (gen-fn0      (gen (caddr expr))))
                 ((symbol? p)        (gen-fn1/rest (gen (caddr expr)) p))
                 ((null? (cdr p))    (gen-fn1      (gen (caddr expr)) (car p)))
                 ((symbol? (cdr p))  (gen-fn2/rest (gen (caddr expr)) (car p) (cdr p)))
                 ((null? (cddr p))   (gen-fn2      (gen (caddr expr)) (car p) (cadr p)))
                 ((symbol? (cddr p)) (gen-fn3/rest (gen (caddr expr)) (car p) (cadr p) (cddr p)))
                 ((null? (cdddr p))  (gen-fn3      (gen (caddr expr)) (car p) (cadr p) (caddr p)))
                 (else               (error "trop de parametres dans une lambda-expression")))))
        ((null? (cdr expr))    (gen-ap0 (gen (car expr))))
        ((null? (cddr expr))   (gen-ap1 (gen (car expr)) (gen (cadr expr))))
        ((null? (cdddr expr))  (gen-ap2 (gen (car expr)) (gen (cadr expr)) (gen (caddr expr))))
        ((null? (cddddr expr)) (gen-ap3 (gen (car expr)) (gen (cadr expr)) (gen (caddr expr))
                                                                           (gen (cadddr expr))))
        (else                  (error "trop d'arguments dans une application de fonction"))))

; --- generation de code pour une constante ---

(define (gen-cst a)
  (lambda (env) a))

; --- generation de code pour une reference de variable ---

(define (gen-ref a)
  (lambda (env) (cdr (assq a env))))

; --- generation de code pour une affectation ---

(define (gen-set a b)
  (lambda (env) (set-cdr! (assq a env) (b env))))

; --- generation de code pour une forme speciale 'if' ---

(define (gen-tst a b c)
  (lambda (env) (if (a env) (b env) (c env))))

; --- generation de code pour une application de fonction ---

(define (gen-ap0 a)
  (lambda (env) ((a env))))

(define (gen-ap1 a b)
  (lambda (env) ((a env) (b env))))

(define (gen-ap2 a b c)
  (lambda (env) ((a env) (b env) (c env))))

(define (gen-ap3 a b c d)
  (lambda (env) ((a env) (b env) (c env) (d env))))

; --- generation de code pour une forme speciale 'lambda' ---

(define (gen-fn0 a)
  (lambda (env)
    (lambda () (a env))))

(define (gen-fn1 a b)
  (lambda (env)
    (lambda (x) (a (cons (cons b x) env)))))

(define (gen-fn2 a b c)
  (lambda (env)
    (lambda (x y) (a (cons (cons b x) (cons (cons c y) env))))))

(define (gen-fn3 a b c d)
  (lambda (env)
    (lambda (x y z) (a (cons (cons b x) (cons (cons c y) (cons (cons d z) env)))))))

(define (gen-fn1/rest a b)
  (lambda (env)
    (lambda x (a (cons (cons b x) env)))))

(define (gen-fn2/rest a b c)
  (lambda (env)
    (lambda (x . y) (a (cons (cons b x) (cons (cons c y) env))))))

(define (gen-fn3/rest a b c d)
  (lambda (env)
    (lambda (x y . z) (a (cons (cons b x) (cons (cons c y) (cons (cons d z) env)))))))

; --- definition de variables globales ---

(define (define-global var val)
  (if (assq var *env-glo*)
    (set-cdr! (assq var *env-glo*) val)
    (begin
      (set-cdr! *env-glo* (cons (car *env-glo*) (cdr *env-glo*)))
      (set-car! *env-glo* (cons var val)))))

(define *env-glo* (list (cons 'define define-global)))
(define-global 'cons  cons )
(define-global 'car   car  )
(define-global 'cdr   cdr  )
(define-global 'null? null?)
(define-global 'not   not  )
(define-global '<     <    )
(define-global '+     +    )
(define-global '-     -    )

; --- pour evaluer, on compile une lambda-expression contenant l'expression a evaluer et ---
; --- on appelle la fonction resultante                                                  ---

(define (evaluer expr)
  ((compile (list 'lambda '() expr))))


;_______________________________________________________________________________________________
;
; Exemples d'utilisation du compilateur:
;
; a) pour evaluer:
;
;      (define fib
;        (lambda (x)
;          (if (< x 2) x (+ (fib (- x 1)) (fib (- x 2))))))
;
;    il faut entrer:
;
       (evaluer '(define 'fib
                   (lambda (x)
                     (if (< x 2) x (+ (fib (- x 1)) (fib (- x 2)))))))
;
; b) pour evaluer:
;
;      (fib 20)
;
;    il faut entrer:
;
       (evaluer '(fib 20))
