;---------------------------------------------------------------------------
;
; Representing 'k'-tuplets of integers by a single integer using a
; 'dove tail' encoding.

(define (inverse f) ; return the inverse of a positive monotonically
                    ; increasing function
  (lambda (y)
    (define (find-power-of-2 x)
      (if (>= (f x) y) x (find-power-of-2 (* x 2))))
    (define (find-inverse x power-of-2)
      (if (= power-of-2 0)
        x
        (let ((new-x (+ x power-of-2)))
          (find-inverse (if (> (f new-x) y) x new-x) (quotient power-of-2 2)))))
    (find-inverse 0 (find-power-of-2 1))))

(define (root k n) ; return 'k'-th root of 'n'
  ((inverse (lambda (x) (expt x k))) n))

(define (fact n) ; return n!
  (if (< n 2) 1 (* (fact (- n 1)) n)))

(define (comb k x) ; return x*(x+1)*...*(x+k-1)/k!
  (do ((i 1 (+ i 1)) (r x (* r (+ x i))))
      ((= i k) (quotient r (fact k)))))

(define (inv-comb k n) ; return x such that (comb k x) <= n < (comb k (+ x 1))
  ((inverse (lambda (x) (comb k x))) n))

(define (sum l) ; return the sum of the values in the list 'l'
  (if (null? l) 0 (+ (car l) (sum (cdr l)))))

;---------------------------------------------------------------------------
;
; Encoding and decoding.
;
; example: (dove-encode 3 '(3 4 0)) => 94  ,  (dove-decode 3 94) => (3 4 0)

(define (dove-encode k l) ; encode list of numbers 'l' of length 'k'
  (if (= k 0) 0 (+ (comb k (sum l)) (dove-encode (- k 1) (cdr l)))))

(define (dove-decode k n) ; decode 'n' into list of numbers of length 'k'
  (if (= k 0)
    '()
    (let* ((x (inv-comb k n)) (l (dove-decode (- k 1) (- n (comb k x)))))
      (cons (- x (sum l)) l))))

;---------------------------------------------------------------------------
;
; Sample use of 'dove-decode':

(define (dove-tail k f) ; call 'f' with all 'k'-tuplets of integers
  (do ((i 0 (+ i 1))) ((= i 200))
    (apply f (dove-decode k i))))

(define (print . l)
  1 ;  (write l) (newline)
)

(dove-tail 3 print) ; print all triplets of integers
; (0 0 0)
; (1 0 0)
; (0 1 0)
; (0 0 1)
; (2 0 0)
; (1 1 0)
; (1 0 1)
; (0 2 0)
; (0 1 1)
; (0 0 2)
; (3 0 0)
; (2 1 0)
; .
; .
; .

(define (sum-of-squares a b)
  (let* ((x (+ a 1)) (y (+ b x)) (z (+ (* x x) (* y y))) (r (root 2 z)))
    (if (= z (expt r 2))
      (begin
;        (display x)
;        (display "^2+")
;        (display y)
;        (display "^2=")
;        (display r)
;        (display "^2")
;        (newline)
1))))

(dove-tail 2 sum-of-squares) ; find all sum of squares
; 3^2+4^2=5^2
; 6^2+8^2=10^2
; 9^2+12^2=15^2
; 5^2+12^2=13^2
; 8^2+15^2=17^2
; 12^2+16^2=20^2
; 15^2+20^2=25^2
; 20^2+21^2=29^2
; 18^2+24^2=30^2
; 10^2+24^2=26^2
; 7^2+24^2=25^2
; 21^2+28^2=35^2
; 16^2+30^2=34^2
; 24^2+32^2=40^2
; 12^2+35^2=37^2
; 27^2+36^2=45^2
; 15^2+36^2=39^2
; 30^2+40^2=50^2
; 9^2+40^2=41^2
; .
; .
; .
